# Create ACE district dummy layers for each NAIP footprint.
# These can be fed into national pooled models.
# This script is also used for making state and rule dummies for prediction points
# (4 M points, PJDs, Navigable Waters)
# Author: Simon Greenhill, sgreenhill@berkeley.edu

library(sf)
library(foreign)
library(glue)
library(data.table)
library(dplyr)
library(pbmcapply)
library(terra)
library(cdlTools)
library(fastDummies)

source('/home/sgreenhill/repos/WOTUS/new_imagery/data/rasterization_functions.R')

split = fread('/home/sgreenhill/wotus/misc/train_test_split.csv')

# encode numerically, going from west to east, snaking north to south
split[, ace_district_numeric := as.numeric(
    factor(ace_district, 
    levels=c('Seattle', 'Portland', 'Walla Walla', 'Sacramento', 'San Francisco',
             'Los Angeles', 'Albuquerque', 'Omaha', 'Kansas City', 'Tulsa', 
             'Ft. Worth', 'Galveston', 'New Orleans', 'Vicksburg', 'Little Rock',
             'Memphis', 'St. Louis', 'Rock Island', 'St. Paul', 'Detroit',
             'Chicago', 'Louisville', 'Nashville', 'Mobile', 'Savannah',
             'Jacksonville', 'Charleston', 'Wilmington', 'Huntington', 
             'Buffalo', 'Pittsburgh', 'Norfolk', 'Baltimore', 'Philadelphia',
             'New York', 'New England')
))]

split[, pointid_str := str_pad(pointid, 6, pad='0')]

# drop duplicates (not sure why these exist, need to investigate)
split = unique(split)

make_district_fe = function(this_pointid, footprints=naip_footprints, 
                            directory='NAIP_ACE_districts', dim=512, overwrite=FALSE) {
    file = glue('/home/sgreenhill/wotus/satellite_imagery/{directory}/{this_pointid}.tif')
    if (file.exists(file) & !overwrite) {
        return(glue('{this_pointid} already run.'))
    }

    # fill a raster with the district number
    footprint = get_footprint(this_pointid, footprints, dim=dim)

    district_val = split[pointid_str==this_pointid, ace_district_numeric]

    footprint = setValues(footprint, district_val)

    writeRaster(
        footprint,
        file,
        overwrite=TRUE
    )

    return(glue('Successfully run {this_pointid}.'))
}

# NAIP footprints
done = pbmclapply(split$pointid_str, make_district_fe, mc.cores=12, ignore.interactive=TRUE)

# Make district FEs for prediction points
make_pred_district_fe = function(
    pointid, 
    value=3, 
    path='/home/sgreenhill/wotus/prediction_points/Sackett_layers/NAIP_ACE_districts',
    skip_done=TRUE
    ) {
    file = glue('{path}/{pointid}.tif')
    if (file.exists(file) & skip_done) {
        return()
    } else if (!dir.exists(path)) {
        dir.create(path, recursive=TRUE)
    }

    # Priest lake is in the Walla Walla regulatory district (id #3 according
    # to the above encoding)
    # (checked manually using QGIS; also confirmed nearby AJD was in WW district)
    # Fun fact: the Walla Walla regulatory district does not include Walla Walla, WA.
    r = raster(nrows=512, ncols=512)
    values(r) = value
    writeRaster(r, file, overwrite=TRUE)
}

# make district FEs for Sackett points
sackett_pts = fread('/home/sgreenhill/wotus/prediction_points/Sackett_sample_NAIP_tiles.csv')
sackett_pts = str_extract(sackett_pts[,prediction_id], '\\d+')
job = pbmclapply(sackett_pts, make_pred_district_fe, mc.cores=8, ignore_interactive=TRUE)


#########################
# 4 M PREDICTION POINTS #
#########################

# make district FEs for all other prediction points
pred_pts = fread('/home/sgreenhill/wotus/prediction_points/prediction_point_metadata.csv')
# merge in the ace district names
ace = st_read('/home/sgreenhill/wotus/admin_boundaries/USACE_Regulatory_Boundary/USACE_Regulatory_Boundary.shp')
cw = as.data.table(ace)[, c('DISTRICT', 'DIST_ABBR')]
pred_pts = merge(pred_pts, cw, by.x='district', by.y='DIST_ABBR')

# encode numerically, going from west to east, snaking north to south 
# note this is the same as above, EXCEPT "Ft. Worth" -> "Fort Worth"
pred_pts[, ace_district_numeric := as.numeric(
    factor(DISTRICT, 
    levels=c('Seattle', 'Portland', 'Walla Walla', 'Sacramento', 'San Francisco',
             'Los Angeles', 'Albuquerque', 'Omaha', 'Kansas City', 'Tulsa', 
             'Fort Worth', 'Galveston', 'New Orleans', 'Vicksburg', 'Little Rock',
             'Memphis', 'St. Louis', 'Rock Island', 'St. Paul', 'Detroit',
             'Chicago', 'Louisville', 'Nashville', 'Mobile', 'Savannah',
             'Jacksonville', 'Charleston', 'Wilmington', 'Huntington', 
             'Buffalo', 'Pittsburgh', 'Norfolk', 'Baltimore', 'Philadelphia',
             'New York', 'New England')
))]

# can use the exact same code for HQ distance, only need to change the value being passed
# do this for the prediction points
print('Starting HQ dist for prediction points')
job = pbmcmapply(
    make_pred_district_fe,
    pointid=pred_pts$prediction_id,
    value=pred_pts$distHQ,
    MoreArgs=list(path='/home/sgreenhill/wotus/prediction_points/grid_points/HQ_dist'),
    mc.cores=20,
    ignore.interactive=TRUE
)
# and the Sackett points
sackett_pts = fread('/home/sgreenhill/wotus/prediction_points/sackett_metadata.csv')
job = pbmcmapply(
    make_pred_district_fe,
    pointid=sackett_pts$id,
    value=sackett_pts$distHQ,
    MoreArgs=list(path='/home/sgreenhill/wotus/prediction_points/Sackett_layers/HQ_dist'),
    mc.cores=30
)

# and we can use the same code for the state FE. Just need to translate from
# FIPS code to the silly custom state code I made
sackett_pts$state_name = fips(sackett_pts$state, to='Name')
# turn this into a numeric code matching create_state_FE.R
sackett_pts[, state_numeric := as.numeric(
    factor(
        state_name,
        levels=c(
            'Washington', 'Oregon', 'California', 'Arizona', 'Nevada', 
            'Utah', 'Idaho', 'Montana', 'Wyoming', 'Colorado', 'New Mexico',
            'Texas', 'Oklahoma', 'Kansas', 'Nebraska', 'South Dakota', 
            'North Dakota', 'Minnesota', 'Iowa', 'Missouri', 'Arkansas', 'Louisiana',
            'Mississippi', 'Tennessee', 'Illinois', 'Wisconsin', 'Michigan',
            'Indiana', 'Kentucky', 'Alabama', 'Florida', 'Georgia', 'South Carolina',
            'North Carolina', 'West Virginia', 'Ohio', 'Pennsylvania', 'New York', 
            'Vermont', 'Maine', 'New Hampshire', 'Massachusetts', 'Connecticut',
            'Rhode Island', 'New Jersey', 'Delaware', 'Maryland', 
            'District of Columbia', 'Virginia'
        )
    )
)
]

job = pbmcmapply(
    make_pred_district_fe,
    pointid=sackett_pts$id,
    value=sackett_pts$state_numeric,
    MoreArgs=list(path='/home/sgreenhill/wotus/prediction_points/Sackett_layers/state',
    mc.cores=30)
)

now make state FEs for the other pred points
print('Starting states for prediction points')
pred_pts$state_name = fips(pred_pts$state, to='Name')
pred_pts[, state_numeric := as.numeric(
    factor(
        state_name,
        levels=c(
            'Washington', 'Oregon', 'California', 'Arizona', 'Nevada', 
            'Utah', 'Idaho', 'Montana', 'Wyoming', 'Colorado', 'New Mexico',
            'Texas', 'Oklahoma', 'Kansas', 'Nebraska', 'South Dakota', 
            'North Dakota', 'Minnesota', 'Iowa', 'Missouri', 'Arkansas', 'Louisiana',
            'Mississippi', 'Tennessee', 'Illinois', 'Wisconsin', 'Michigan',
            'Indiana', 'Kentucky', 'Alabama', 'Florida', 'Georgia', 'South Carolina',
            'North Carolina', 'West Virginia', 'Ohio', 'Pennsylvania', 'New York', 
            'Vermont', 'Maine', 'New Hampshire', 'Massachusetts', 'Connecticut',
            'Rhode Island', 'New Jersey', 'Delaware', 'Maryland', 
            'District of Columbia', 'Virginia'
        )
    )
)
]

job = pbmcmapply(
    make_pred_district_fe,
    pointid=pred_pts$prediction_id,
    value=pred_pts$state_numeric,
    MoreArgs=list(path='/home/sgreenhill/wotus/prediction_points/grid_points/state'),
    ignore.interactive=TRUE,
    mc.cores=20
)

# Finally, make Rule FEs
# Sackett
print('Starting Sackett Rules')
sackett_pts = fread('/home/sgreenhill/wotus/prediction_points/Sackett_sample_NAIP_tiles.csv')

# Rapanos
job = pbmclapply(
    sackett_pts$prediction_id,
    make_pred_district_fe,
    value=1,
    path='/home/sgreenhill/wotus/prediction_points/Sackett_layers/rules/Rapanos/',
    mc.cores=20,
    ignore.interactive=TRUE
)
# CWR
job = pbmclapply(
    sackett_pts$prediction_id,
    make_pred_district_fe,
    value=2,
    path='/home/sgreenhill/wotus/prediction_points/Sackett_layers/rules/CWR/',
    mc.cores=20,
    ignore.interactive=TRUE
)
# NWPR
job = pbmclapply(
    sackett_pts$prediction_id,
    make_pred_district_fe,
    value=3,
    path='/home/sgreenhill/wotus/prediction_points/Sackett_layers/rules/NWPR/',
    mc.cores=20,
    ignore.interactive=TRUE
)

# Grid points
print('Starting rules for grid points')
job = pbmclapply(
    pred_pts$prediction_id,
    make_pred_district_fe,
    value=1,
    path='/home/sgreenhill/wotus/prediction_points/grid_points/rules/Rapanos/',
    mc.cores=30,
    ignore.interactive=TRUE
)
# CWR
job = pbmclapply(
    pred_pts$prediction_id,
    make_pred_district_fe,
    value=2,
    path='/home/sgreenhill/wotus/prediction_points/grid_points/rules/CWR/',
    mc.cores=30,
    ignore.interactive=TRUE
)
# NWPR
job = pbmclapply(
    pred_pts$prediction_id,
    make_pred_district_fe,
    value=3,
    path='/home/sgreenhill/wotus/prediction_points/grid_points/rules/NWPR/',
    mc.cores=30,
    ignore.interactive=TRUE
)

########
# PJDs #
########

# use metadata provided by Hannah
pjds = fread('/home/sgreenhill/wotus/prediction_points/PJDs/pjd_metadata.csv')


# Rule FEs
##########

# Rapanos
job = pbmclapply(
    pjds$prediction_id,
    make_pred_district_fe,
    value=1,
    path='/home/sgreenhill/wotus/prediction_points/PJDs/rules/Rapanos/',
    mc.cores=90,
    ignore.interactive=TRUE
)

# CWR
job = pbmclapply(
    pjds$prediction_id,
    make_pred_district_fe,
    value=2,
    path='/home/sgreenhill/wotus/prediction_points/PJDs/rules/CWR/',
    mc.cores=90,
    ignore.interactive=TRUE
)

# NWPR
job = pbmclapply(
    pjds$prediction_id,
    make_pred_district_fe,
    value=3,
    path='/home/sgreenhill/wotus/prediction_points/PJDs/rules/NWPR/',
    mc.cores=90,
    ignore.interactive=TRUE
)

# State FEs
###########

pjds$state_name = fips(pjds$state, to='Name')
pjds[, state_numeric := as.numeric(
    factor(
        state_name,
        levels=c(
            'Washington', 'Oregon', 'California', 'Arizona', 'Nevada', 
            'Utah', 'Idaho', 'Montana', 'Wyoming', 'Colorado', 'New Mexico',
            'Texas', 'Oklahoma', 'Kansas', 'Nebraska', 'South Dakota', 
            'North Dakota', 'Minnesota', 'Iowa', 'Missouri', 'Arkansas', 'Louisiana',
            'Mississippi', 'Tennessee', 'Illinois', 'Wisconsin', 'Michigan',
            'Indiana', 'Kentucky', 'Alabama', 'Florida', 'Georgia', 'South Carolina',
            'North Carolina', 'West Virginia', 'Ohio', 'Pennsylvania', 'New York', 
            'Vermont', 'Maine', 'New Hampshire', 'Massachusetts', 'Connecticut',
            'Rhode Island', 'New Jersey', 'Delaware', 'Maryland', 
            'District of Columbia', 'Virginia'
        )
    )
)
]

job = pbmcmapply(
    make_pred_district_fe,
    pointid=pjds$prediction_id,
    value=pjds$state_numeric,
    MoreArgs=list(path='/home/sgreenhill/wotus/prediction_points/PJDs/state'),
    ignore.interactive=TRUE,
    mc.cores=90
)


# HQ Distance
#############

pjd_meta = fread('/home/sgreenhill/wotus/prediction_points/PJDs/pjd_metadata.csv') %>%
    dplyr::select(district, distHQ, prediction_id)
pjds = merge(pjds, pjd_meta)

job = pbmcmapply(
    make_pred_district_fe,
    pointid=pjds$prediction_id,
    value=pjds$distHQ,
    MoreArgs=list(path='/home/sgreenhill/wotus/prediction_points/PJDs/HQ_dist'),
    ignore.interactive=TRUE,
    mc.cores=90
)

# District FE
#############

ace = st_read('/home/sgreenhill/wotus/admin_boundaries/USACE_Regulatory_Boundary/USACE_Regulatory_Boundary.shp')
cw = as.data.table(ace)[, c('DISTRICT', 'DIST_ABBR')]
pjds = merge(pjds, cw, by.x='district', by.y='DIST_ABBR')

# encode numerically, going from west to east, snaking north to south 
pjds[, ace_district_numeric := as.numeric(
    factor(DISTRICT, 
    levels=c('Seattle', 'Portland', 'Walla Walla', 'Sacramento', 'San Francisco',
             'Los Angeles', 'Albuquerque', 'Omaha', 'Kansas City', 'Tulsa', 
             'Fort Worth', 'Galveston', 'New Orleans', 'Vicksburg', 'Little Rock',
             'Memphis', 'St. Louis', 'Rock Island', 'St. Paul', 'Detroit',
             'Chicago', 'Louisville', 'Nashville', 'Mobile', 'Savannah',
             'Jacksonville', 'Charleston', 'Wilmington', 'Huntington', 
             'Buffalo', 'Pittsburgh', 'Norfolk', 'Baltimore', 'Philadelphia',
             'New York', 'New England')
))]

job = pbmcmapply(
    make_pred_district_fe,
    pointid=pjds$prediction_id,
    value=pjds$ace_district_numeric,
    MoreArgs=list(path='/home/sgreenhill/wotus/prediction_points/PJDs/ACE_districts', skip_done=FALSE),
    mc.cores=32
)

####################
# NAVIGABLE WATERS #
####################

nav_path = '/home/sgreenhill/wotus/prediction_points/nav_water'
nav_ids = fread(glue('{nav_water}/nav_water_metadata.csv'))

# Rule FEs
##########
# Rapanos
job = pbmclapply(
    nav_ids$prediction_id, make_pred_district_fe,
    value=1,
    path=glue('{nav_path}/rules/Rapanos'),
    mc.cores=32,
    ignore.interactive=TRUE
)

# CWR
job = pbmclapply(
    nav_ids$prediction_id, make_pred_district_fe,
    value=2,
    path=glue('{nav_path}/rules/CWR'),
    mc.cores=32,
    ignore.interactive=TRUE
)

# NWPR
job = pbmclapply(
    nav_ids$prediction_id, make_pred_district_fe,
    value=3,
    path=glue('{nav_path}/rules/NWPR'),
    mc.cores=32,
    ignore.interactive=TRUE
)

# State FEs
###########

nav_ids$state_name = fips(nav_ids$state, to='Name')
nav_ids[, state_numeric := as.numeric(
    factor(
        state_name,
        levels=c(
            'Washington', 'Oregon', 'California', 'Arizona', 'Nevada', 
            'Utah', 'Idaho', 'Montana', 'Wyoming', 'Colorado', 'New Mexico',
            'Texas', 'Oklahoma', 'Kansas', 'Nebraska', 'South Dakota', 
            'North Dakota', 'Minnesota', 'Iowa', 'Missouri', 'Arkansas', 'Louisiana',
            'Mississippi', 'Tennessee', 'Illinois', 'Wisconsin', 'Michigan',
            'Indiana', 'Kentucky', 'Alabama', 'Florida', 'Georgia', 'South Carolina',
            'North Carolina', 'West Virginia', 'Ohio', 'Pennsylvania', 'New York', 
            'Vermont', 'Maine', 'New Hampshire', 'Massachusetts', 'Connecticut',
            'Rhode Island', 'New Jersey', 'Delaware', 'Maryland', 
            'District of Columbia', 'Virginia'
        )
    )
)
]

job = pbmcmapply(
    make_pred_district_fe,
    pointid=nav_ids$prediction_id,
    value=nav_ids$state_numeric,
    MoreArgs=list(path='/home/sgreenhill/wotus/prediction_points/nav_water/state'),
    ignore.interactive=TRUE,
    mc.cores=32
)


# HQ Distance
#############

job = pbmcmapply(
    make_pred_district_fe,
    pointid=nav_ids$prediction_id,
    value=nav_ids$distHQ,
    MoreArgs=list(path='/home/sgreenhill/wotus/prediction_points/nav_water/HQ_dist'),
    ignore.interactive=TRUE,
    mc.cores=32
)

# District FE
#############

ace = st_read('/home/sgreenhill/wotus/admin_boundaries/USACE_Regulatory_Boundary/USACE_Regulatory_Boundary.shp')
cw = as.data.table(ace)[, c('DISTRICT', 'DIST_ABBR')]
nav_ids = merge(nav_ids, cw, by.x='district', by.y='DIST_ABBR')

# encode numerically, going from west to east, snaking north to south 
# note this is the same as above, EXCEPT "Ft. Worth" -> "Fort Worth"
nav_ids[, ace_district_numeric := as.numeric(
    factor(DISTRICT, 
    levels=c('Seattle', 'Portland', 'Walla Walla', 'Sacramento', 'San Francisco',
             'Los Angeles', 'Albuquerque', 'Omaha', 'Kansas City', 'Tulsa', 
             'Fort Worth', 'Galveston', 'New Orleans', 'Vicksburg', 'Little Rock',
             'Memphis', 'St. Louis', 'Rock Island', 'St. Paul', 'Detroit',
             'Chicago', 'Louisville', 'Nashville', 'Mobile', 'Savannah',
             'Jacksonville', 'Charleston', 'Wilmington', 'Huntington', 
             'Buffalo', 'Pittsburgh', 'Norfolk', 'Baltimore', 'Philadelphia',
             'New York', 'New England')
))]

job = pbmcmapply(
    make_pred_district_fe,
    pointid=nav_ids$prediction_id,
    value=nav_ids$ace_district_numeric,
    MoreArgs=list(path='/home/sgreenhill/wotus/prediction_points/nav_water/ACE_districts', skip_done=FALSE),
    mc.cores=32
)