import os
import multiprocessing as mp
import numpy as np
import pandas as pd
from tqdm import tqdm
import pickle
from skimage.io import imread
from functools import partial

spath = '/home/sgreenhill/wotus/satellite_imagery'

n_workers = int(mp.cpu_count())

def load_image(file, image_directory, n_layers=None, select_layers=None):
    '''
    n_layers: None or int. If int, will make sure the axis with size n_layers
        is in the last position, to make the mean be taken by layer.
    select_layers: None or list. If None, all layers of the last axis are used.
        If list, keep only the layers of the last axis whose indices are in this
        list.
    '''
    try:
        if file[-3:] == 'npy':
            image = np.load(f'{image_directory}/{file}')
        else:
            image = imread(f'{image_directory}/{file}')
    except Exception:
        print(file)
    
    if n_layers:
        shp = image.shape
        # Identify the axis that has size n_layers
        i = np.argwhere([shp[i] == n_layers] for i in range(len(shp)))
        # Make this axis the last axis
        image = np.moveaxis(image, i, -1)

    if select_layers:
        image = image[:, :, select_layers]
    return image


def image_mean(file, image_directory, **kwargs):
    image = load_image(file, image_directory, **kwargs)
    try:
        mu = np.mean(image, axis=(0, 1))
    except Exception:
        print(file)
    return mu


def image_variance(file, image_directory, means, **kwargs):
    image = load_image(file, image_directory, **kwargs)
    return np.mean(np.square(image - means), axis=(0, 1))


def parallel_mean(image_directory, files, workers=n_workers, **kwargs):
    with mp.Pool(workers) as p:
        res = np.vstack(list(
            tqdm(p.imap(partial(image_mean, image_directory=image_directory, **kwargs), files), 
            total=len(files))))
    return np.mean(res, keepdims=True, axis=0)


def parallel_variance(image_directory, files, means, workers=n_workers, **kwargs):
    with mp.Pool(workers) as p:
        res = np.vstack(list(tqdm(
            p.imap(
                partial(image_variance, image_directory=image_directory,
                        means=means, **kwargs),
                files
            ),
            total=len(files)
        )))
    return np.mean(res, keepdims=True, axis=0)


def image_check(file, image_directory, n_layers):
    '''
    Diagnostic to identify ill-formatted Sentinel-2 images
    Example problem pointid: 108796
    '''
    image = load_image(file, image_directory, n_layers=n_layers)
    if image.shape != (256, 256, 11):
        res_dict = {
            'file': [file],
            'shape': [image.shape],
            'all0': [(image == 0).all()]
        }
        df = pd.DataFrame(res_dict)
        return df


# S2 drops
check_df = pd.read_csv('/home/sgreenhill/wotus/diagnostics/SentinelMissingData.csv')
s2_drops = np.load('/home/sgreenhill/wotus/misc/s2_crossborder_and_offshore_pointids.p.npy',
                    allow_pickle=True).astype(str)
# NAIP drops
naip_drops = np.load('/home/sgreenhill/wotus/misc/crossborder_and_offshore_pointids.p', 
                allow_pickle=True)
# pad string
naip_drops = [str(f).rjust(6, '0') for f in naip_drops]

all_drops = check_df.pointid_str.tolist() + naip_drops + list(s2_drops)

def exclude_drops(files, prefix, suffix, missings=all_drops):
    '''
    Given a list of files, drop the ones where S2 data are missing
    Args:
        files: List of str. The list of files to check over
        prefix: str. Everything before the pointid in the filenames
        suffix: str. Everything after the pointid in the filenames
        missings: list-like. The pointids that are missings.
    '''
    missings = [prefix + str(p) + suffix for p in missings]
    keeps = list(set(files) - set(missings))
    return keeps


def calculate_mean_sd(slug, directory, prefix, suffix,
                      save_directory=f'{spath}/layer_mean_sd',
                      drop_s2=False,
                      **kwargs):
    '''
    Helper function for calculating the mean and standard deviation of a layer
    and saving it.
    Args:
        slug: str. The short name for saving this layer
        directory: str. The directory where these data live.
        prefix: str. Everything before the pointid in the filenames
        suffix: str. Everything after the pointid in the filenames
        save_directory: str. Where you want to save the output.
        drop_s2: bool. Drop files where S2 data are missing?
    '''
    if drop_s2:
        files = exclude_drops(os.listdir(directory), prefix, suffix)
    else:
        files = os.listdir(directory)
    
    print('Calculating mean(s)...')
    means = parallel_mean(directory, files, **kwargs)
    print('Calculating variance(s)...')
    variances = parallel_variance(directory, files, means=means, **kwargs)
    sds = np.sqrt(variances)

    results = {'means': means, 'sds': sds}
    
    savepath = f'{save_directory}/{slug}_mean_std.p'
    
    with open(savepath, 'wb') as f:
        pickle.dump(results, f)

    print(f'Saved results at {savepath}')
    return results


def check_load(p, folder='ACE_districts', 
               path='/home/sgreenhill/wotus/prediction_points/grid_points/',
               error_on_missing=False,
               prefix='',
               extension='.tif'
               ):
    '''
    Checks that a file can be loaded without error
    args:
        error_on_missing: bool. Should a filename be returned if the corresponding
            file can't be located?
    '''
    try:
        imread(f'{path}/{folder}/{prefix}{p}{extension}')
    # missing files handled separately
    except FileNotFoundError:
        if error_on_missing:
            return np.array([p])
        else:
            return np.array([])
    except:
        return np.array([p])
    return np.array([])

