# Common functions for rasterizing shapefiles to NAIP tiles
library(ggplot2)
library(stringr)
library(fasterize)
library(raster)
library(reticulate)
np = import("numpy")
library(RStoolbox)
library(patchwork)
library(readr)
library(data.table)
library(sf)
library(glue)
theme_set(theme_void())
# do not save auxiliary files with raster outputs 
rgdal::setCPLConfigOption("GDAL_PAM_ENABLED", "FALSE")

footprints = '/home/sgreenhill/wotus/satellite_imagery/tile_footprints'

naip_footprints = st_read(glue('{footprints}/NAIP_tiles.shp'))
st_crs(naip_footprints) = 4326

s2_footprints = st_read(glue('{footprints}/Sentinel-2_tiles.shp'))
st_crs(s2_footprints) = 4326

ajds = fread('/home/sgreenhill/wotus/AJD/jds_wet_dry_season_clean_v2.csv')
ajds[, pointid_str := str_pad(pointid, 6, pad='0')]

get_footprint = function(this_pointid, footprints=naip_footprints, dim=512) {
	#' Function to get a 512x512 raster of a NAIP image's footprint
	footprint = footprints %>% dplyr::filter(pointid==this_pointid)
	footprint_r = fasterize::raster(footprint, nrows=dim, ncols=dim)

	return(footprint_r)
}

rasterize_shapefile = function(
    pointid, shpfile, footprint, field, rasterize_fn=fasterize,
	na_val=0, save=FALSE, save_path=NULL
) {
	# rasterize the shapefile
	rasterized = rasterize_fn(shpfile, footprint, field=field,
	                     background=na_val
						 )

	# set crs
	crs(rasterized) = "+proj=longlat"

	if (save) {
		if (is.null(save_path)) {
			print('No save path given, cannot save.')
		} else {
			dir.create(save_path, showWarnings=FALSE, recursive=TRUE)
			fname = glue('{save_path}/{pointid}.tif')
			# If the raster already exists, this means the NAIP footprint overlaps
			# two or more shapefiles. If this is the case, load the existing one 
			# and average it with the new one, with handling for NAs
			if (file.exists(fname)) {
				old_raster = raster(fname)
				stacked = raster::stack(list(rasterized, old_raster))

				for (i in 1:nlayers(stacked)) {
					# replace the na_vals with NAs for averaging
					stacked[[i]][stacked[[i]] == na_val] = NA
				}
				rasterized = raster::calc(stacked, mean, na.rm=TRUE)
				# re-replace the NAs with the desired na_val
				rasterized[is.na(rasterized)] = na_val
				
				# save a list of these pointids, for later inspection
				df = data.frame('pointid'=c(pointid))
				fname2 = glue('{save_path}/multiple_pointids.csv')

				if (file.exists(fname2)) {
					write_csv(df, fname2, append=TRUE, col_names=FALSE)
				} else {
					write_csv(df, fname2)
				}
			}
			writeRaster(rasterized, fname, overwrite=TRUE)
		}
	}
	return(rasterized)
}

three_panel_diagnostic = function(
    point, shpfile, field, raster, footprints, save_dir, raster_processor,
	state=NULL, categorical=TRUE
    ) {
	tic = Sys.time()
	# plot the raster
	# df = as.data.frame(raster, xy=TRUE) %>%
	# 	mutate(layer=as.factor(layer)) # TODO: abstract from this issue
	df = raster_processor(raster)

	raster_plot = ggplot(df) +
		geom_raster(aes(x=x, y=y, fill=layer)) +
		labs(title='Rasterized') +
		theme(
			legend.position = 'bottom',
			legend.title = element_blank()
		)

	if (categorical) {
		raster_plot = raster_plot +
			scale_fill_brewer(palette='BuPu') +
			guides(
				fill = guide_legend(nrow=3, byrow=TRUE)
			)
	} else {
		raster_plot = raster_plot +
			scale_fill_distiller(palette='BuPu')
	}
	

	# plot the shapefile
	footprint = get_naip_footprint(point, footprints)
	ext = extent(footprint)

	# need to fix any messed up geometries before cropping
	xmin = ext[1]
	xmax = ext[2]
	ymin = ext[3]
	ymax = ext[4]
	shp_plot = ggplot(shpfile) +
		geom_sf(aes(fill=field)) +
		scale_fill_brewer(palette='PuBuGn') +
		lims(
			x = c(xmin, xmax),
			y = c(ymin, ymax)
		) +
		labs(title='Shapefile') +
		guides(
			fill = guide_legend(nrow=3, byrow=TRUE)
		) +
		theme(
			'legend.position' = 'bottom',
			'legend.title' = element_blank()
		)

	# load and plot the image
	img = np$load(glue('/home/sgreenhill/wotus/satellite_imagery/NAIP/',
	                   'NAIP_{point}.npy'))

	# make the RBG channels into a raster for easier plotting
	img_rasts = apply(img[, , 1:3], 3, raster::raster)
	img_rast = raster::stack(img_rasts)

	img_plot = ggRGB(img_rast, r=1, g=2, b=3) +
		labs(title = 'NAIP image')

	# make a 3-panel plot
	wotus = ajds[pointid_str == point, wotus]
	title = ifelse(is.null(state),
	 			   glue("PointID: {point}, WOTUS: {wotus}"),
				   glue("PointID: {point}, state: {state}, WOTUS: {wotus}")
	)
	combined_plot = img_plot + shp_plot + raster_plot + 
		plot_annotation(title)

	# save out
	ggsave(glue('{save_dir}{point}.png'),
	       width=15, height=8)

	toc = Sys.time()

	print(toc - tic)
}


n_panel_diagnostic = function(
    point, rasters, save_dir=NULL, raster_processors,
	raster_names, # categoricals, 
	state=NULL, plot_shp=TRUE, plot_NIR=FALSE, shpfile=NULL, footprints=NULL
    ) {
	# plot the raster
	# df = as.data.frame(raster, xy=TRUE) %>%
	# 	mutate(layer=as.factor(layer)) # TODO: abstract from this issue
	raster_plots = c()
	for (i in seq_along(rasters)) {
		# categorical = categoricals[i]
		df = raster_processors[i][[1]](rasters[i][[1]])

		raster_plot = ggplot(df) +
			geom_raster(aes(x=x, y=y, fill=layer)) +
			labs(title=glue('{raster_names[i]}')) +
			theme(
				legend.position = 'bottom',
				legend.title = element_blank()
			)

		raster_plot = raster_plot +
				scale_fill_brewer(palette='YlGnBu') +
				guides(
					fill = guide_legend(nrow=3, byrow=TRUE)
				)

		# if (categorical) {
		# 	raster_plot = raster_plot +
		# 		scale_fill_brewer(palette='BuPu') +
		# 		guides(
		# 			fill = guide_legend(nrow=3, byrow=TRUE)
		# 		)
		# } else {
		# 	raster_plot = raster_plot +
		# 		scale_fill_distiller(palette='BuPu')
		# }

		raster_plots = 	c(raster_plots, list(raster_plot))
	}

	if (!is.null(shpfile)) {
		# plot the shapefile
		footprint = get_naip_footprint(point, footprints)
		ext = extent(footprint)
		xmin = ext[1]
		xmax = ext[2]
		ymin = ext[3]
		ymax = ext[4]
		shp_plot = ggplot(shpfile) +
			geom_sf() +
			scale_fill_brewer(palette='PuBuGn') +
			lims(
				x = c(xmin, xmax),
				y = c(ymin, ymax)
			) +
			labs(title='Shapefile') +
			guides(
				fill = guide_legend(nrow=3, byrow=TRUE)
			) +
			theme(
				'legend.position' = 'bottom',
				'legend.title' = element_blank()
			)
	}
	
	# load and plot the image
	img = np$load(glue('/home/sgreenhill/wotus/satellite_imagery/NAIP/',
	                   'NAIP_{point}.npy'))

	# make the RBG channels into a raster for easier plotting
	img_rasts = apply(img[, , 1:3], 3, raster::raster)
	img_rast = raster::stack(img_rasts)

	img_plot = ggRGB(img_rast, r=1, g=2, b=3) +
		labs(title = 'NAIP RGB')

	if (plot_NIR) {
		NIR = raster(img[, , 4])
		NIR_plot = ggR(NIR) +
			labs(title = 'NAIP NIR')
	}

	# make the combined plot
	wotus = ajds[pointid_str == point, wotus]
	title = ifelse(is.null(state),
	 			   glue("PointID: {point}, WOTUS: {wotus}"),
				   glue("PointID: {point}, state: {state}, WOTUS: {wotus}")
	)

	to_combine = c(list(img_plot), raster_plots)
	counter = 1
	if (plot_NIR) {
		to_combine = append(to_combine, list(NIR_plot), after=1)
		counter = counter + 1
	}
	if (!is.null(shpfile)) {
		to_combine = append(to_combine, list(shp_plot), after=counter)
	}

	combined_plot = Reduce(`+`, c(list(img_plot), list(NIR_plot), raster_plots)) + # Reduce(`+`, to_combine) +
		plot_annotation(title) +
		plot_layout(nrow=1)

	# save out 
	if (!is.null(save_dir)) {
		ggsave(glue('{save_dir}/{point}.png'), combined_plot,
	       width=15, height=8)
	} else {
		return(combined_plot)
	}
	
}

# Raster processors: NHD
process_StreamOrde_raster = function(rast) {
    #' Function to turn a rasterized StreamOrder into a df that can be plotted
    df = as.data.frame(rast, xy=TRUE) %>%
		dplyr::rename(layer=names(rast)) %>%
        dplyr::mutate(layer=factor(layer, levels=seq(-1, 7)))
    
    return(df)
}

process_numeric_raster_to_fact = function(rast) {
    df = as.data.frame(rast, xy=TRUE) %>%
		dplyr::rename(layer=names(rast)) %>%
        dplyr::mutate(layer=as.factor(round(layer, digits=1)))
    return(df)
}

process_numeric_raster = function(rast, na.val) {
	df = as.data.frame(rast, xy=TRUE) %>%
		dplyr::rename(layer=names(rast)) %>%
		 dplyr::mutate(layer = ifelse(layer == na.val, NA, layer))

	return(df)
}

process_numeric_raster_nhd = function(rast, na.val) {
    df = as.data.frame(rast, xy=TRUE) %>%
		dplyr::rename(layer=names(rast)) %>%
        dplyr::mutate(layer=as.factor(round(layer, digits=1))) %>%
		dplyr::mutate(layer=dplyr::recode_factor(layer, `-1`='Not in NHD'))
    return(df)
}

process_Fcode_raster = function(rast) {
    df = as.data.frame(rast, xy=TRUE) %>%
		dplyr::rename(layer=names(rast)) %>%
        dplyr::mutate(
            layer = dplyr::recode(layer,
                `1` = "Artificial path",
                `2` = "Canal/ditch",
                `3` = "Coastline",
                `4` = "Connector",
                `5` = "Pipeline",
                `6` = "Intermittent/ephemeral stream",
                `7` = "Perennial stream",
                `8` = "Other stream",
                `9` = "Underground conduit",
                `0` = "Other",
                `-1` = "Not in NHD"
            )
        )
    return(df)
}

