'''
Helper functions for loading data for deep learning models.
'''

import os
import pickle
import numpy as np
import pandas as pd
import torch
from torch.utils.data import Dataset, DataLoader
from torchvision import transforms
import skimage.io as skio
from dataclasses import dataclass
from glob import glob
from tqdm import tqdm

spath = '/home/sgreenhill/wotus/satellite_imagery/'

def make_split_dicts(split, wotus=True, cowardin=False, resource=False, ajd=False,
                     verbose=True):
    '''
    Takes in a pd.DataFrame with the train/test split info, returns
    dictionaries of file names and labels for each split fold.

    args:
        split: pd.DataFrame. The train test split.
        wotus: bool. Return wotus dicts?
        cowardin: bool or "simple". If simple, loads the simplified cowardin
            code (just stream, wetland, or other). If true, loads the 16-class
            code. If False, does not load cowarding codes.
        resource: bool. If True, loads resource types. Otherwise, does not.
        ajd: bool. Return AJD dicts?
        verbose: bool. Print proportion WOTUS in each split? Will error if 
            wotus=False.
    '''
    # get the file names of the images in each fold
    names = {'train': [], 'val': [], 'test': []}
    labels = {'train': [], 'val': [], 'test': []}
    cowardin_labels = {'train': [], 'val': [], 'test': []}
    resource_labels = {'train': [], 'val': [], 'test': []}
    ajd_labels = {'train': [], 'val': [], 'test': []}

    image_dir=f'{spath}/NAIP'
    prefix = 'NAIP_'

    # missings = pd.DataFrame({'pointid': []})

    for i, row in split.iterrows():
        fold = row['split_group']

        # separate handling for the AJD and prediction point data here
        if ajd:
            if row.id[:3] == 'ajd':
                pointid = row.id[4:]
                name = f'{prefix}{pointid}.npy'
            else:
                pointid = row.id[5:]
                name = int(pointid)
        else:
            pointid = str(row['pointid']).zfill(6)
            name = f'{prefix}{pointid}.npy'
            
        names[fold].append(name)
        if wotus:
            if row['wotus'] == 'Yes':
                labels[fold].append(1)
            else:
                labels[fold].append(0)
        if cowardin == 'simple':
            cowardin_labels[fold].append(row['cowardin_simple_numeric'])
        elif cowardin:
            cowardin_labels[fold].append(row['cowardin_numeric'])
        if resource:
            resource_labels[fold].append(row['resource_numeric'])
        if ajd:
            ajd_labels[fold].append(row['ajd'])
    
    if verbose:
        # print split diagnostics
        for fold in labels.keys():
            if wotus:
                print(f'Prop. of {fold} labels WOTUS: {sum(labels[fold]) / len(labels[fold]):0.2f}.')
                print(f'Number of {fold} observations: {len(labels[fold])}')
            if ajd:
                print(f'Prop. of {fold} labels AJD: {sum(ajd_labels[fold]) / len(ajd_labels[fold]):0.2f}.')
                print(f'Number of {fold} observations: {len(ajd_labels[fold])}')

    return names, labels, cowardin_labels, resource_labels, ajd_labels


def drop_missings(split, drop_naip=True, drop_S2=False):
    '''
    Drop missing training data, mostly locations that are offshore.
    Parameters:
        split (pd.DataFrame): split df.
        drop_overseas (bool): Drop pointids where the NAIP footprint spands
            international borders or is offshore?
        drop_S2 (bool): Drop pointids where the Sentinel-2 data is missing?
    Returns:
        pd.DataFrame: the split df with drops removed
    '''
    dir = '/home/sgreenhill/wotus/misc'
    if drop_naip:
        # naip overseas footprints
        os_d = np.load(f'{dir}/crossborder_and_offshore_pointids.p',
                       allow_pickle=True)

        # missing distance to HQ
        n_d = np.load(f'{dir}/naip_dist_to_hq_missing.npy')

        # misisng district info
        n_dd = np.load(f'{dir}/naip_ace_district_missing.npy')

        os_d = np.concatenate([os_d, n_d, n_dd])
        split = split.loc[~split.pointid.isin(os_d)]
    if drop_S2:
        # s2
        os_s2_d  = np.load(f'{dir}/crossborder_and_offshore_pointids_s2.npy')

        # Get the ones where the imagery is missing
        s2_d = pd.read_csv('/home/sgreenhill/wotus/diagnostics/SentinelMissingData.csv')
        # get the pointids
        s2_d_p = s2_d.pointid_str.astype(int).values

        # get the ones where PRISM is missing (see data/identify_missing_PRISM.ipynb)
        s2_d2 = np.load(f'{dir}/s2_crossborder_and_offshore_pointids.p.npy', allow_pickle=True)

        # get the ones where distance to HQ is missing
        s2_d3 = np.load(f'{dir}/s2_dist_to_hq_missing.npy')
        
        s2_d_p = np.concatenate([os_s2_d, s2_d_p, s2_d2, s2_d3])
        split = split.loc[~split.pointid.isin(s2_d_p)]

    return split

def augment(ar):
    '''
    Helper function for implementing data augmentations
    '''

    # rotation augmentation 
    k = np.random.randint(low=0, high=4)
    ar = np.rot90(ar, k, axes=(1,2))

    # flip augmentation
    flip1 = np.random.uniform()
    flip2 = np.random.uniform()
    if flip1 > 0.5:
        ar = np.flip(ar, axis=1)
    if flip2 > 0.5:
        ar = np.flip(ar, axis=2)

    return ar
    

def normalize(image, norm_dict_path, drops=None):
    '''
    Helper function for normalizing data given a dictionary containing the
    overall mean and standard deviation for the set of data.

    Args:
    norm_dict_path: dictionary with keys 'means' and 'stds', containing the means
        and standard errors to be used in the normalization
    drops: list-like of indices to drop from means and stds. 
        Default None (keep all). Useful for when you are not using all of this
        image's layers.
    '''

    image = torch.from_numpy(image.copy())

    norm_dict = np.load(norm_dict_path, allow_pickle=True)
    if drops:
        norm_dict = {k: np.delete(norm_dict[k], drops) for k in norm_dict.keys()}

    n_layers = image.shape[0]

    mu, sigma = [np.reshape(norm_dict[k], (n_layers,)) for k in norm_dict.keys()]

    image = transforms.Normalize(mu, sigma)(image)

    return image



def load_file(path, image_size=512):
    '''
    Helper function for loading rasterized data saved as .tif files in a format
    that can be immediately concatenated with the image
    
    Args:
        path: str. The path to load from.
    '''
    ext = path[-4:]
    # if extension is not specified, guess it
    if ext[0] != '.':
        match = glob(path + '*')
        if len(match) > 1:
            raise ValueError(f'More than one file at {path}')
        elif len(match) == 0:
            raise FileNotFoundError(f'No file found at {path}')

        path = match[0]
    
    if ext in ['.npz', '.npy']:
        im = np.load(path, allow_pickle=True)
        if ext == '.npz':
            im = im['arr_0']

    if path[-4:] == '.npy':
        im = np.load(path, allow_pickle=True)
    else:
        im = skio.imread(path)

    im = im.astype('float32')

    # If needed, expand dims so the array is three-dimensional for concatenation
    # with other 3-D arrays.
    if len(im.shape) == 2:
        im = np.expand_dims(im, axis=0)

    for i, ax in enumerate(im.shape):
        # it looks like for some prediction points we accidentally 
        # grabbed data of the wrong size. Drop the 1 extra pixel.
        if ax > image_size:
            if abs(ax - image_size) > 10:
                raise ValueError('Layer size is off by more than 10 pixels')
            
            im = im.take(indices=range(0, image_size), axis=i)
    
    return im


class NAIPDataset(Dataset):
    
    def __init__(
        self, image_names, fold, labels=None, image_size=512, n_naip_bands=4,
        augment=True, normalize=True, 
        normalize_path=f'{spath}/layer_mean_sd',
        root_dir=spath,
        naip=True,
        image_dir='AJD_inputs/NAIP',
        image_prefix=None,
        nwi=False,
        nwi_dir='AJD_inputs/NWI/',
        nwi_prefix='NWI_',
        nhd=False,
        nhd_stacked=True,
        nhd_dir='AJD_inputs/NHD',
        nhd_prefix='NHD_',
        dem=False,
        dem_dir='DEM',
        dem_prefix='DEM_',
        ecoregion=False,
        ecoregion_dir='Ecoregion',
        ecoregion_prefix='Ecoregion_',
        nlcd=False,
        nlcd_dir='NLCD',
        nlcd_prefix='NLCD_',
        prism=False,
        prism_dir='PRISM',
        prism_prefix='PRISM_',
        prism_ppt_dir = 'PRISM_ppt',
        prism_ppt_prefix='PRISM_ppt_',
        gnatsgo=False,
        gnatsgo_dir='gNATSGO',
        gnatsgo_stacked=False,
        gnatsgo_prefix=None,
        district_dummies=False,
        district_dummies_dir='ACE_districts',
        district_dummies_prefix='',
        hq_dist=False,
        hq_dist_dir='HQ_dist',
        hq_dist_prefix='',
        regrid=False,
        regrid_dir='REGRID',
        rule_dummies=False,
        rule_dummies_dir='NAIP_rule_dummies',
        rule_dummies_prefix='',
        states=False,
        states_dir='state',
        states_prefix='',
        subset_layers=False,
        save=False,
        load_saved=False,
        save_dir=None,
        cowardin=False,
        cowardin_labels=None,
        resource=False,
        resource_labels=None,
        ajd=False,
        ajd_labels=None,
        wotus=True,
        prediction_group=None,
        rule_one_hot=False,
        permute_layer=None
    ):
        '''
        args:
        All self-explanatory booleans for the data layers and strings for the
        directories. Partial documentation for less intuitive args below.
        cowardin: bool or "simple". If simple, loads the simplified cowardin
            code (just stream, wetland, or other). If true, loads the 16-class
            code. If False, does not load cowarding codes.
        resource: bool. If True, loads resource types. Otherwise, does not.
        subset_layers: bool. For each data source, subset only to the layers
            that plausibly have the greatest effect on WOTUS status?
        rule_one_hot: bool. If True, encodes the rule variables as one layer 
            per rule with 1s and 0s instead of a single layer with integers
        permute_layer: int, list, or None. Layer(s) to permute. Used for feature
            importance.
        '''
        self.image_dir = image_dir
        self.image_prefix = image_prefix
        self.image_names = image_names
        self.labels = labels
        self.image_size = image_size
        self.n_naip_bands = n_naip_bands
        self.fold = fold
        self.transform = transforms.Compose([transforms.ToTensor()])
        self.augment = augment
        self.normalize = normalize
        self.normalize_path = normalize_path
        self.naip = naip
        self.nwi = nwi
        self.nwi_prefix = nwi_prefix
        self.root_dir = root_dir
        self.nwi_dir = nwi_dir
        self.nwi_prefix=nwi_prefix
        self.nhd = nhd
        self.nhd_stacked = nhd_stacked
        self.nhd_prefix = nhd_prefix
        self.nhd_dir = nhd_dir
        self.dem = dem
        self.dem_dir = dem_dir
        self.dem_prefix = dem_prefix
        self.ecoregion = ecoregion
        self.ecoregion_dir = ecoregion_dir
        self.ecoregion_prefix = ecoregion_prefix
        self.nlcd = nlcd
        self.nlcd_dir = nlcd_dir
        self.nlcd_prefix = nlcd_prefix
        self.prism = prism
        self.prism_dir = prism_dir
        self.prism_prefix = prism_prefix
        self.prism_ppt_dir = prism_ppt_dir
        self.prism_ppt_prefix = prism_ppt_prefix
        self.gnatsgo = gnatsgo
        self.gnatsgo_dir = gnatsgo_dir
        self.gnatsgo_stacked = gnatsgo_stacked
        self.gnatsgo_prefix = gnatsgo_prefix
        self.district_dummies = district_dummies
        self.district_dummies_dir = district_dummies_dir
        self.district_dummies_prefix = district_dummies_prefix
        self.hq_dist = hq_dist
        self.regrid = regrid
        self.regrid_dir = regrid_dir
        self.hq_dist_dir = hq_dist_dir
        self.hq_dist_prefix = hq_dist_prefix
        self.rule_dummies = rule_dummies
        self.rule_dummies_dir = rule_dummies_dir
        self.rule_dummies_prefix = rule_dummies_prefix
        self.states = states
        self.states_dir = states_dir
        self.states_prefix = states_prefix
        self.subset_layers = subset_layers
        self.save = save
        self.load_saved = load_saved
        self.save_dir = save_dir
        self.wotus = wotus
        self.cowardin = cowardin
        self.resource = resource
        self.cowardin_labels = cowardin_labels
        self.resource_labels = resource_labels
        self.ajd = ajd
        self.ajd_labels = ajd_labels
        self.prediction_group = prediction_group
        self.rule_one_hot = rule_one_hot
        self.permute_layer = permute_layer
        
    def __getitem__(self, item):
        name = self.image_names[item]
        folder = self.image_dir
        
        # this step only needed for NAIP naming convention, not for Sackett prediction points
        if isinstance(name, str):
            if len(name) > 4 and '.' in name:
                pointid_str = name[5:-4]
            else:
                pointid_str = name
                # special handling for the NAIP naming conventions for different
                # prediction points
                if self.prediction_group == 'Sackett':
                    name = f'NAIP_Sackett_{pointid_str}.tiff'
                # elif self.prediction_group == 'PJD':
                #     name = f'PJD_{pointid_str}.tiff'
            pointid = int(pointid_str)
        else:
            pointid_str = str(name)
            pointid = name

        if self.labels or self.ajd_labels:
            if self.wotus:
                label = self.labels[item]
            if bool(self.cowardin):
                cowardin = self.cowardin_labels[item]
            if self.resource:
                resource = self.resource_labels[item]
            if self.ajd:
                ajd_label = self.ajd_labels[item]

        
        # load the saved version if it exists
        loaded=False
        if self.save_dir:
            os.makedirs(self.save_dir, exist_ok=True)
            save_fpath = f'{self.save_dir}/{pointid_str}.npz'
            save_dict = f'{self.save_dir}/data_dict.p'
        
        if self.load_saved and os.path.exists(save_fpath) and os.path.exists(save_dict):
            # There's a few reasons this can fail. Namely, if there's a mismatch
            # between the dictionaries or if the image file is corrupted 
            # (due to the process being interrupted)
            try:
                image = load_file(save_fpath)

                # check consistency and drop unwanted layers
                data_dict = np.load(save_dict, allow_pickle=True)
                assert data_dict['augment'] == self.augment, "Augment mismatch"
                assert data_dict['normalize'] == self.normalize, "Normalize mismatch"
                drops = []
                bools = [
                    self.naip,
                    self.nwi, self.nhd, self.dem, self.ecoregion, self.nlcd, 
                    self.prism, self.gnatsgo, self.regrid, self.district_dummies,
                    self.rule_dummies, self.hq_dist, self.states
                ]
                keys = [
                    'naip_ids',
                    'nwi_id', 'nhd_ids', 'dem_id', 'ecoregion_id', 'nlcd_id',
                    'prism_ids', 'gnatsgo_ids', 'regrid_id', 'district_dummies_id',
                    'rule_dummies_id', 'hq_dist_id', 'state_id'
                ]

                for b, key in zip(bools, keys):
                    # special handling for the NAIP bands
                    if key == 'naip_ids' and len(data_dict['naip_ids']) != self.n_naip_bands:
                        # the only band we would be dropping from NAIP is NIR,
                        # which is the 4th band
                        drops += [data_dict['naip_ids'][-1]]
                    
                    if b:
                        assert data_dict[key], f'No {key}'
                    elif key in data_dict:
                        drops += data_dict[key]
                
                # implement drops
                image = np.delete(image, drops, axis=0)

                loaded = True
            except:
                print(f'Tried and failed to load data for {item}.')
                loaded = False
            
        # if we didn't successfully load the file, or we didn't want to, 
        # load from the raw data.
        if not loaded:
            # track layer ids for the data dictionary
            layer_id = 0
            data_dict = {}

            # set up paths to grab means and SDs from for nomralization
            if self.normalize:
                normalize_path = self.normalize_path
            # try:
            if self.naip:
                # for the grid prediction points, NAIP is the only one where 
                # we always need it to be a 7-character string; all others enter
                # as ints.
                # if isinstance(name, int):
                #     naip_name = str(name).ljust(7,'0') + '.tiff'
                # else:
                #     naip_name = name
                try:
                    image = load_file(f'{self.root_dir}/{folder}/{pointid_str}.')
                except (FileNotFoundError, ValueError):
                    # naip_name = name

                    # load the NAIP image
                    if self.image_prefix is None:
                        image = load_file(f'{self.root_dir}/{folder}/{pointid_str}.')
                    else:
                        image = load_file(f'{self.root_dir}/{folder}/{self.image_prefix}{pointid_str}.')

                # If more than self.n_naip_bands bands, keep only the first self.n_naip_bands
                # this allows us to use RGB only if NIR is not desired.
                image = np.moveaxis(image, -1, 0)
                image = image[:self.n_naip_bands]

                # Normalize the raw image data
                if self.normalize:
                    if self.n_naip_bands < 4:
                        drops = np.arange(self.n_naip_bands, 4)
                    else: 
                        drops = None
                    
                    image = normalize(image, f'{normalize_path}/NAIP_mean_std.p', drops=drops)
                
                data_dict['naip_ids'] = list(np.arange(layer_id, layer_id + self.n_naip_bands))
                layer_id += self.n_naip_bands

            else:
                # initialize a 0-dimensional array for other layers to be
                # concatenated to
                image = np.empty((0, self.image_size, self.image_size))

            # load rasterized NWI
            if self.nwi:
                nwi = load_file(f'{self.root_dir}/{self.nwi_dir}/{self.nwi_prefix}{pointid_str}.')
                for i, ax in enumerate(nwi.shape):
                    # it looks like for some prediction points we accidentally 
                    # grabbed data of the wrong size. Drop the 1 extra pixel.
                    if ax > self.image_size:
                        nwi = nwi.take(indices=range(0, self.image_size), axis=i)

                data_dict['nwi_id'] = [layer_id]
                layer_id += 1
                image = np.concatenate([image, nwi], axis=0)
                

            # load rasterized NHD
            if self.nhd:
                if self.nhd_stacked:
                    # layer order: COMID, fcode, path length, stream order, low flow,
                    # high flow
                    nhd_data = load_file(f'{self.root_dir}/{self.nhd_dir}/{self.nhd_prefix}{pointid_str}.')
                    
                    # make the first axis the axis with the different layers
                    nhd_data = np.moveaxis(nhd_data, -1, 0)

                    # drop comid
                    nhd_data = np.delete(nhd_data, 0, axis=0)

                    # re-order so that the layers that need to be normalized using
                    # pre-saved mean and sd are consecutive (i.e., make stream order)
                    # the 2nd layer instead of the 3rd
                    nhd_data = nhd_data[(0, 2, 1, 3, 4), :, :]

                    # Normalize fcode, pathlength, lowflow, high flow
                    # note that Fcode is normalized in stacked version and not
                    if self.normalize: 
                        # normalize fcode by dividing by 10,000 
                        # (natively it's a 5-digit integer)
                        nhd_data[0] = nhd_data[0] / 1e5

                        # use the saved mean and SD for pathlength, lowflow, high flow
                        # pathlength
                        nhd_data[2:] = normalize(nhd_data[2:], f'{normalize_path}/NAIP_NHD_mean_std.p')
                                
                    image = np.concatenate([image, nhd_data], axis=0)
                    
                    n_nhd = nhd_data.shape[0]
                    data_dict['nhd_ids'] = list(np.arange(layer_id, layer_id + n_nhd))
                    layer_id += n_nhd
                else:
                    nhd_var_list = ['Fcode', 'Pathlength', 'hiflow', 'lowflow', 
                                'StreamOrde']

                    nhd_data = [
                        load_file(f'{self.root_dir}/{self.nhd_dir}/{dir}/{pointid_str}.tif')
                        for dir in nhd_var_list
                    ]

                    # Normalize Pathlength, hiflow, lowflow
                    if self.normalize:
                        for i, v in enumerate(nhd_var_list):
                            if v in ['Pathlength', 'hiflow', 'lowflow']:
                                nhd_data[i] = normalize(
                                    nhd_data[i],
                                    f'{normalize_path}/NAIP_NHD_{v}_mean_std.p'
                                )
                    
                    nhd = np.concatenate(nhd_data, axis=0)
                    image = np.concatenate([image, nhd], axis=0)
                    n_nhd = len(nhd_var_list)
                    data_dict['nhd_ids'] = list(np.arange(layer_id, layer_id + n_nhd))
                    layer_id += n_nhd    
            # load DEM
            if self.dem:
                dem = load_file(f'{self.root_dir}/{self.dem_dir}/{self.dem_prefix}{pointid_str}.tiff')
                if self.normalize:
                    dem = normalize(
                        dem, 
                        f'{normalize_path}/DEM_mean_std.p'
                    )

                image = np.concatenate([image, dem], axis=0)
                data_dict['dem_id'] = [layer_id]
                layer_id += 1
            

            # load ecoregions
            if self.ecoregion:
                eco = load_file(f'{self.root_dir}/{self.ecoregion_dir}/{self.ecoregion_prefix}{pointid_str}.tiff')
                # divide by 100 to make mean and std more similar to other layers
                # natively these data are three-digit numbers
                if self.normalize:
                    eco = eco / 100
                image = np.concatenate([image, eco], axis=0)
                data_dict['ecoregion_id'] = [layer_id]
                layer_id += 1

            # load nlcd
            if self.nlcd:
                nlcd = load_file(f'{self.root_dir}/{self.nlcd_dir}/{self.nlcd_prefix}{pointid_str}.tiff')
                # divide by 10 to make mean and std more similar to other layers
                # natively these data are 2-digit numbers
                if self.normalize:
                    nlcd = nlcd / 10

                for i, ax in enumerate(nlcd.shape):
                    # it looks like for some prediction points we accidentally 
                    # grabbed data of the wrong size. Drop the 1 extra pixel.
                    if ax > self.image_size:
                        # print(f'{pointid_str}: Resizing NLCD axis {i} from {ax} to {self.image_size}')
                        nlcd = nlcd.take(indices=range(0, self.image_size), axis=i)

                image = np.concatenate([image, nlcd], axis=0)
                data_dict['nlcd_id'] = [layer_id]
                layer_id += 1

            # load PRISM
            if self.prism:
                prism = load_file(f'{self.root_dir}/{self.prism_dir}/{self.prism_prefix}{pointid_str}.tiff')

                # move the axis with all the layers to be in the 0th position
                prism = np.moveaxis(prism, -1, 0)

                # drop the DEM (because we already have another one separately)
                prism = np.delete(prism, 7, axis=0)

                # there is 1 pointid with missing values (141341).
                # this is because it cross the border between New York and Canada.
                # The missing values are -10^38, replace these with -50 before normalizing
                prism[prism < -50] = -50

                if self.normalize:
                    prism = normalize(
                        prism,
                        f'{normalize_path}/PRISM_mean_std.p',
                        drops=[7]
                    )

                # precip is in a separate file and thus needs to be handled separately
                prism_prcp = load_file(f'{self.root_dir}/{self.prism_ppt_dir}/{self.prism_ppt_prefix}{pointid_str}.tiff')

                # replace missings, as above
                prism_prcp[prism_prcp < -50] = -50
                if self.normalize:
                    prism_prcp = normalize(
                        prism_prcp, f'{normalize_path}/PRISM_ppt_mean_std.p'
                    )

                prism = np.concatenate([prism, prism_prcp], axis=0)

                if self.subset_layers:
                    # keep only tdmean (daily mean temp) and precip
                    prism = prism[[3, 9]]
                image = np.concatenate([image, prism], axis=0)

                n_prism = prism.shape[0]
                data_dict['prism_ids'] = list(np.arange(layer_id, layer_id + n_prism))
                layer_id += n_prism

            # load gNATSGO
            if self.gnatsgo:
                if self.gnatsgo_stacked:
                    gnatsgo_stacked = load_file(f'{self.root_dir}/{self.gnatsgo_dir}/{self.gnatsgo_prefix}{pointid_str}.tiff')
                    # order of stacked files: Tax, Pond, HydrcRating, Flood, depth to water table
                    # make these match the order in gnatsgo_var_list below
                    gnatsgo_data = [
                        np.expand_dims(gnatsgo_stacked[:, :, i], 0) 
                        for i in [3, 2, 1, 0, 4]
                        ]
                else:
                    gnatsgo_var_list = [
                        'Flood/gNATSGO_DCD', 'HydrcRatng/gNATSGO_HydrcRatng', 
                        'Pond/gNATSGO_Pond', 'Tax/gNATSGO_Tax', 'WTD/gNATSGO_WTD']

                    gnatsgo_data = [
                        load_file(f'{self.root_dir}/{self.gnatsgo_dir}_{dir}_{pointid_str}.tiff')
                        for dir in gnatsgo_var_list
                    ]

                if self.normalize:
                    # Normalize WTD
                    gnatsgo_data[4] = normalize(
                        gnatsgo_data[4], f'{normalize_path}/gNATSGO_WTD_mean_std.p'
                    )

                    # Divide HydrcRatng by 10 to make mean and std more similar
                    # to other layers (natively it is 0-100%)
                    gnatsgo_data[1] = gnatsgo_data[1] / 10

                    # Divide soil taxonomy by 1000 to make mean and std more similar
                    # to other layers (natively it is integer 0-18000)
                    gnatsgo_data[3] = gnatsgo_data[3] / 1000

                    # Note ponding and flooding frequency are take values in 
                    # integers <10, so don't need to normalize them further

                gnatsgo = np.concatenate(gnatsgo_data, axis=0)

                if self.subset_layers:
                    gnatsgo = gnatsgo[[2, 4]]

                image = np.concatenate([image, gnatsgo], axis=0)

                n_gnatsgo = gnatsgo.shape[0]
                data_dict['gnatsgo_ids'] = list(np.arange(layer_id, layer_id + n_gnatsgo))
                layer_id += n_gnatsgo

            # load parcels
            if self.regrid:
                regrid_data = load_file(f'{self.root_dir}/{self.regrid_dir}/REGRID_{pointid_str}.tiff')
                # no normalization needed--0 or 1 values natively
                image = np.concatenate([image, regrid_data], axis=0)

                data_dict['regrid_id'] = [layer_id]
                layer_id += 1


            # augment all the data that needs to be augmented
            if self.fold == 'train' and self.augment:
                image = augment(image)

            # load district dummies if desired (note these do not need to be 
            # augmented)
            if self.district_dummies:
                try:
                    district = load_file(
                        f'{self.root_dir}/{self.district_dummies_dir}/{self.district_dummies_prefix}{pointid_str}.tif'
                    )
                except FileNotFoundError:
                    district = load_file(
                        f'{self.root_dir}/{self.district_dummies_dir}/{self.district_dummies_prefix}{pointid}.tif'
                    )
                # divide by 10 to make mean and std more similar to other
                # layers (these are normally integers 1-40 or so)
                if self.normalize:
                    district = district / 10
                data_dict['district_dummies_id'] = [layer_id]
                layer_id += 1
                
                image = np.concatenate([image, district], axis=0)


            # load rule dummies (again, no need for augmentation)
            if self.rule_dummies:
                rule = load_file(f'{self.root_dir}/{self.rule_dummies_dir}/{self.rule_dummies_prefix}{pointid_str}.tif')
                if self.rule_one_hot:
                    value = int(np.unique(rule))
                    n_rules = 3
                    rule = np.zeros(
                        shape=(n_rules, self.image_size, self.image_size),
                        dtype=np.float32
                    )
                    rule[value - 1, :, :] = 1
                    data_dict['rule_dummies_id'] = list(np.arange(layer_id, layer_id + n_rules))
                    layer_id += n_rules
                else:
                    data_dict['rule_dummies_id'] = [layer_id]
                    layer_id += 1
                image = np.concatenate([image, rule], axis=0)
                

            # load hq distance (also does not need to be augmented)
            if self.hq_dist:
                try:
                    naip_dist = load_file(f'{self.root_dir}/{self.hq_dist_dir}/{self.hq_dist_prefix}{pointid_str}')
                except (FileNotFoundError, ValueError):
                    naip_dist = load_file(f'{self.root_dir}/{self.hq_dist_dir}/{self.hq_dist_prefix}{pointid}.')
                if self.normalize:
                    naip_dist = normalize(naip_dist, f'{normalize_path}/NAIP_dist_to_hq_mean_std.p')
                image = np.concatenate([image, naip_dist], axis=0)

                data_dict['hq_dist_id'] = [layer_id]
                layer_id += 1

            # load state (also does not need to be augmented)
            if self.states:
                try:
                    state = load_file(f'{self.root_dir}/{self.states_dir}/{self.states_prefix}{pointid_str}.tif')
                except FileNotFoundError:
                    state = load_file(f'{self.root_dir}/{self.states_dir}/{self.states_prefix}{pointid}.tif')

                data_dict['state_id'] = [layer_id]
                layer_id += 1
                
                image = np.concatenate([image, state], axis=0)

            if self.save:
                data_dict['augment'] = self.augment
                data_dict['normalize'] = self.normalize
                # if data dictionary already saved, check equivalence
                data_dict_file = f'{self.save_dir}/data_dict.p'
                if os.path.exists(data_dict_file):
                    old_data_dict = np.load(data_dict_file, allow_pickle=True)
                    assert old_data_dict == data_dict, "Data dict mismatch"
                else:
                    if not os.path.exists(self.save_dir):
                        os.mkdir(self.save_dir)
                    with open(data_dict_file, 'wb') as f:
                        pickle.dump(data_dict, f)
                    # np.save(data_dict_file, data_dict, allow_pickle=True)
                np.savez_compressed(save_fpath, image)

        if self.permute_layer is not None:
            # permute layer(s)
            random_pointid_str = np.random.choice(self.image_names)[5:-4]
            random_fpath = f'{self.save_dir}/{random_pointid_str}.npz'
            random_image = load_file(random_fpath)

            image[self.permute_layer] = random_image[self.permute_layer]
        # pass to torch
        image = torch.from_numpy(image.copy())

        ret = {'image': image, 'pointid': np.array([pointid])}
        if self.labels or self.ajd_labels:
            if self.wotus:
                ret['wotus'] = np.array([label])
            if self.ajd:
                ret['ajd'] = np.array([ajd_label])
            # using bool(.) nests cases cowardin=='simple' and cowardin==True
            if bool(self.cowardin): 
                ret['cowardin'] = np.array([cowardin])
            if self.resource:
                ret['resource'] = np.array([resource])
            
        return ret
    

    def __len__(self):
        return len(self.image_names)


def get_NAIP_dataloaders(
    names, labels, batch_size, num_cores, n_naip_bands, naip, nwi, nhd, dem, 
    ecoregion, nlcd, prism, gnatsgo, district_dummies, hq_dist, regrid,
    rule_dummies, states,
    augment=True, normalize=True, normalize_path=f'{spath}/layer_mean_sd', return_datasets=False, subset_layers=False,
    return_config=False, save=False, load_saved=False, save_dir=False,
    shuffle=True, root_dir=spath, image_dir='AJD_inputs/NAIP', image_prefix=None,
    nwi_dir='AJD_inputs/NWI', nhd_dir='AJD_inputs/NHD', 
    nhd_stacked=True, nhd_prefix='NHD_',
    nwi_prefix='NWI_', dem_dir='DEM',
    ecoregion_dir='Ecoregion', nlcd_dir='NLCD', prism_dir='PRISM', 
    prism_ppt_dir='PRISM_ppt', gnatsgo_dir='gNATSGO', gnatsgo_stacked=False,
    district_dummies_dir='ACE_districts', district_dummies_prefix='',
    hq_dist_dir='HQ_dist', hq_dist_prefix='',
    regrid_dir='REGRID', rule_dummies_dir='NAIP_rule_dummies',
    rule_dummies_prefix='', states_dir='state', states_prefix='',
    dem_prefix='DEM_', ecoregion_prefix='Ecoregion_', nlcd_prefix='NLCD_',
    prism_prefix='PRISM_', prism_ppt_prefix='PRISM_ppt_', gnatsgo_prefix=None,
    folds=['train', 'val'], cowardin=False, resource=False, cowardin_labels=None,
    resource_labels=None, prediction_group=None, rule_one_hot=False,
    ajd=False, ajd_labels=None, wotus=True, permute_layer=None):
    '''
    Helper function to get the dataloaders. Arguments as those for __init__
    method of NAIPDataset class.
    '''
    datasets = {x: NAIPDataset(image_names=names[x],labels=labels[x], fold=x,
                               n_naip_bands=n_naip_bands, normalize=normalize,
                               normalize_path=normalize_path, 
                               augment=augment, root_dir=root_dir,
                               naip=naip, nwi=nwi, nhd=nhd, dem=dem, 
                               ecoregion=ecoregion, nlcd=nlcd, prism=prism,
                               gnatsgo=gnatsgo, district_dummies=district_dummies,
                               hq_dist=hq_dist,
                               regrid=regrid,
                               subset_layers=subset_layers,
                               image_dir=image_dir,
                               image_prefix=image_prefix,
                               nwi_dir=nwi_dir,
                               nwi_prefix=nwi_prefix,
                               nhd_dir=nhd_dir,
                               nhd_stacked=nhd_stacked,
                               nhd_prefix=nhd_prefix,
                               dem_dir=dem_dir,
                               ecoregion_dir=ecoregion_dir,
                               nlcd_dir=nlcd_dir,
                               prism_dir=prism_dir,
                               prism_ppt_dir=prism_ppt_dir,
                               gnatsgo_dir=gnatsgo_dir,
                               gnatsgo_stacked=gnatsgo_stacked,
                               district_dummies_dir=district_dummies_dir,
                               district_dummies_prefix=district_dummies_prefix,
                               hq_dist_dir=hq_dist_dir,
                               hq_dist_prefix=hq_dist_prefix,
                               regrid_dir=regrid_dir,
                               dem_prefix=dem_prefix,
                               ecoregion_prefix=ecoregion_prefix,
                               nlcd_prefix=nlcd_prefix,
                               prism_prefix=prism_prefix,
                               prism_ppt_prefix=prism_ppt_prefix,
                               gnatsgo_prefix=gnatsgo_prefix,
                               rule_dummies=rule_dummies,
                               rule_dummies_dir=rule_dummies_dir,
                               rule_dummies_prefix=rule_dummies_prefix,
                               states=states,
                               states_dir=states_dir,
                               states_prefix=states_prefix,
                               save=save,
                               load_saved=load_saved,
                               save_dir=save_dir,
                               cowardin=cowardin,
                               resource=resource,
                               cowardin_labels=cowardin_labels[x],
                               resource_labels=resource_labels[x],
                               ajd=ajd,
                               ajd_labels=ajd_labels[x],
                               wotus=wotus,
                               prediction_group=prediction_group,
                               rule_one_hot=rule_one_hot,
                               permute_layer=permute_layer
                               )
            for x in folds}

    dataloaders = {x: DataLoader(datasets[x],
                                batch_size=batch_size,
                                shuffle=shuffle,
                                num_workers=num_cores,
                                pin_memory=False)
                for x in folds}

    ret = [dataloaders]

    if return_datasets:
        ret = [datasets, dataloaders]
    if return_config:
        # dictionary tracking the inputs and settings for the data
        if naip:
            naip_n_bands = 'naip' + str(n_naip_bands)
        else:
            naip_n_bands = 'naip'

        config = {
            'batch_size': batch_size, 'augmented': augment, 'normalized': normalize,
            'shuffled': shuffle, 'subset_layers': subset_layers,
            naip_n_bands: naip_n_bands,
            'nwi': nwi, 'nhd': nhd, 'dem': dem, 'ecoregion': ecoregion, 
            'nlcd': nlcd, 'prism': prism, 'gnatsgo': gnatsgo, 
            'district_dummies': district_dummies, 'hq_dist': hq_dist,
            'regrid': regrid, 'rule_dummies': rule_dummies, 
            'rule_one_hot': rule_one_hot, 'states': states,
            'cowardin': cowardin==True, 'cowardin_simple': cowardin=='simple', 
            'resource': resource, 'ajd': ajd
        }

        ret += [config]

    return ret