'''
Functions for plotting model inputs and outputs
'''

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import sys
import os

from diagnostics_utils import group_accuracy

def plot_metrics(train_metrics, val_metrics, train_preds, val_preds,
                 fs=(20, 10), MA=None, by_epoch=True, model_dir=None, title=None):
    fig, ax = plt.subplots(figsize=fs, ncols=2)

    num_epochs = max(train_metrics.epoch) + 1

    # calculate accuracy
    if by_epoch:
        grouper = 'epoch'
    else:
        grouper = 'batch' 

    train_acc = pd.DataFrame(
        train_preds.groupby(grouper).apply(group_accuracy),
        columns=['accuracy']
    )
    val_acc = pd.DataFrame(
        val_preds.groupby(grouper).apply(group_accuracy),
        columns=['accuracy']
    )

    if by_epoch:
        train_metrics =  train_metrics.groupby('epoch').mean()
        val_metrics = val_metrics.groupby('epoch').mean()
    if MA:
        for df in [train_metrics, val_metrics]:
            df['loss_MA'] = df.loss.rolling(window=MA).mean()
        train_loss_toplot = train_metrics.loss_MA
        val_loss_toplot = val_metrics.loss_MA

        for df in [train_acc, val_acc]:
            df['accuracy_MA'] = df.accuracy.rolling(window=MA).mean()
        train_acc_toplot = train_acc.accuracy_MA
        val_acc_toplot = val_acc.accuracy_MA
    else:
        train_loss_toplot = train_metrics.loss
        val_loss_toplot = val_metrics.loss
        train_acc_toplot = train_acc.accuracy
        val_acc_toplot = val_acc.accuracy

    ax[0].plot(train_metrics.index, train_loss_toplot, color='tab:blue', label='Train')
    ax[0].set_title(f'Loss')
    ax[0].set_ylabel('Loss')

    ax[0].plot(val_metrics.index, val_loss_toplot, color='tab:orange', label='Val')

    ax[1].plot(train_acc.index, train_acc_toplot, color='tab:blue')
    ax[1].plot(val_acc.index, val_acc_toplot, color='tab:orange')
    ax[1].set_title('Accuracy')
    ax[1].set_title('Accuracy')

    plt.suptitle(title)
    if by_epoch:
        [a.set_xlabel('Epoch') for a in ax]
        ma_fill = 'epoch'
    else:
        [a.set_xlabel('Batch number') for a in ax]
        ma_fill = 'batch'
    fig.legend()
    # plt.tight_layout()
    
    if MA:
        ax[0].annotate(f'{MA}-{ma_fill} moving average plotted', (0, -0.1),
            xycoords='axes fraction')

    if model_dir:
        plt.savefig(f'{model_dir}/accuracy_loss_plot_epoch.pdf')

    plt.show()


def plot_training_curves(
    models,
    model_names,
    root='/home/sgreenhill/repos/WOTUS/new_imagery/experiments/',
    MA=3,
    linestyles=['solid', 'dotted', 'dashed'],
    prefixes = [''],
    colors=None,
    xlim=None
    ):

    train_metrics = [pd.read_csv(f'{root}{m}/train_metrics.csv').groupby('epoch').mean() for m in models]
    val_metrics = [pd.read_csv(f'{root}{m}/val_metrics.csv').groupby('epoch').mean() for m in models]
    train_preds = [pd.read_csv(f'{root}{m}/train_preds.csv') for m in models]
    val_preds = [pd.read_csv(f'{root}{m}/val_preds.csv') for m in models]

    from pdb import set_trace; set_trace()
    for p in prefixes:
        train_acc = [
            pd.DataFrame(
                t.groupby('epoch').apply(
                    group_accuracy,
                    preds_col=f'{p}predictions',
                    labels_col=f'{p}labels'), 
                columns=['accuracy'])
            for t in train_preds
        ]

        val_acc = [
            pd.DataFrame(
                t.groupby('epoch').apply(
                    group_accuracy, 
                    preds_col=f'{p}predictions',
                    labels_col=f'{p}labels')
                , columns=['accuracy']) 
            for t in val_preds
            ]

        fig, ax = plt.subplots(figsize=(20,10), ncols=2)
        for i in range(len(models)):
            if colors:
                val_col = colors[i]['val']
                train_col = colors[i]['train']
            else:
                val_col = 'tab:orange'
                train_col = 'tab:blue'
            
            index = train_metrics[i].index
            train_loss_toplot = train_metrics[i].loss.rolling(window=MA).mean()
            val_loss_toplot = val_metrics[i].loss.rolling(window=MA).mean()
            train_acc_toplot = train_acc[i].accuracy.rolling(window=MA).mean()
            val_acc_toplot = val_acc[i].accuracy.rolling(window=MA).mean()
            ax[0].plot(index, train_loss_toplot, color=train_col, label=model_names[i] + ', train', linestyle=linestyles[i])
            ax[0].set_title(f'Loss')
            ax[0].set_ylabel('Loss')

            ax[0].plot(index, val_loss_toplot, color=val_col, label=model_names[i] + ', val', linestyle=linestyles[i])

            ax[1].plot(index, train_acc_toplot, color=train_col, linestyle=linestyles[i])
            ax[1].plot(index, val_acc_toplot, color=val_col, linestyle=linestyles[i])
            ax[1].set_title('Accuracy')
            ax[1].set_title('Accuracy')

            if xlim:
                ax[0].set_xlim(0, xlim)
                ax[1].set_xlim(0, xlim)

        fig.legend()
        plt.show()



def imshow(data, printdf=False):
    """Imshow for Tensor."""
    inp = data['image']
    classes = data['wotus']
    files = data['file']
    
    nc = 4
    n_images = inp.shape[0]
    nr = int(np.ceil(n_images / nc))
    fig, ax = plt.subplots(figsize=(20, 20), nrows=nr, ncols=nc)
    axs = ax.ravel()
    inp = inp.numpy().transpose(0, 2, 3, 1).astype(np.uint8)
    
    for i in range(n_images):
        axs[i].imshow(inp[i, :, :, :])
        # axs[i].set_title(class_names[int(classes[i])])
        # get the lat lon
        pointid = int(files[i][5:-4])
        ll = ajds.loc[ajds.pointid == pointid, ['latitude', 'longitude']]
        axs[i].annotate(f'Pointid: {pointid}, lat: {ll.latitude.values}, lon: {ll.longitude.values}', 
                        (0, -0.1), xycoords='axes fraction')
        
        if printdf:
            display(ajds.loc[ajds.pointid==pointid])
        
    
    plt.show()

def imshow_from_ds(ids, fold, datasets, printdf=False,
    layer_titles=None,
    rgb_together=True,
    save=None,
    multi_row=True
):
    """Imshow for Tensor."""
    
    # if plot_nwi:
    #     n_images = len(ids) * 2
    # else:
    
    if rgb_together:
        start_layer = 3
        column_adjust = 2
    else:
        start_layer = 0
        column_adjust = 0
    
    # set up number of rows and columns
    data = datasets[fold].__getitem__(ids[0])
    inp = data['image'].numpy()
    
    n_layers = inp.shape[0]
    n_images = len(ids)
    nc = n_layers - column_adjust
    nr = n_images

    fig = plt.figure(figsize=(50, 5))
    gs = GridSpec(nr, nc)

    for i, id in enumerate(ids):
        try:
            data = datasets[fold].__getitem__(id)
        except Exception as e:
            print('Error, skipping')
            print(e)
            plt.close()
            return

        inp = data['image'].numpy()

        pointid = data['pointid'][0]

        print(pointid)
            

        if rgb_together:
            image = inp[:3]
            image = image.transpose(1, 2, 0).astype(np.uint8)
            
            ax = fig.add_subplot(gs[i, 0])
            rgb = ax.imshow(image)

            fig.colorbar(rgb, ax=ax, orientation='horizontal', pad=0.01)

            if layer_titles is not None:
                ax.set_title(layer_titles[0])

        plt.axis('off')
        for j in range(start_layer, n_layers):
            this_ax = fig.add_subplot(gs[i, j-column_adjust])
            l = this_ax.imshow(inp[j])
            plt.axis('off')
            # colorbar for numeric non-image stuff:
            # if j > 3:
            fig.colorbar(l, ax=this_ax, orientation='horizontal', pad=0.01)

            if layer_titles is not None:
                this_ax.set_title(layer_titles[j])

        
        if printdf:
            display(ajds.loc[ajds.pointid==pointid])
        
    gs.tight_layout(fig)
    plt.subplots_adjust(hspace=0)

    if save:
        plt.savefig(save, dpi=1000, bbox_inches='tight', pad_inches=0)

    plt.show()

class_names = {0: 'Not WOTUS', 1: 'Yes WOTUS'}


def points_to_bin(x, y, vals, scale=.1, coastlines_shp=None, plot_crs=None,
                  binary=False):
    """bins points over 2d space with bin sizes specified by scale
        args:
            x,y: nx1 arrays of locations in 1 dimension each
            vals: nx1 array of values to be averaged
            scale: the edge of a bin/box in {x,y} units.
            coastlines_shp: GeoPandas DataFrame. Shapefile of the coastlines for the
                region you want to plot.
            plot_crs: NOT CURRENTLY IMPLEMENTED.
                the CRS you want your plot in. This argument has not been
                tested thoroughly and should be used with caution. Your best bet
                is to make sure your coastlines shapefile is in the same CRS as
                the CRS of the plot you are trying to create before feeding the
                shapefile into this function.
            binary: bool. Round values to whole numbers? In cases where vals are in 
                  [0, 1], this leads to rounded values that are binary.
        returns:
            x0, y0: kx1, mx1 arrays of the x and y gridpoints
            vals_grid: (m-1)x(k-1) resulting aggregated values
        """
    x_range = np.max(x) - np.min(x)
    y_range = np.max(y) - np.min(y)
    bin_shapes = [int(y_range / scale), int(x_range / scale)]

    sums_grid, y0, x0 = np.histogram2d(y, x, bins=bin_shapes, weights=vals)

    counts, y1, x1 = np.histogram2d(y, x, bins=bin_shapes)

    vals_grid = sums_grid / counts
    
    vals_grid = np.ma.masked_invalid(vals_grid)

    if binary:
        vals_grid = np.round(vals_grid,0)
    
    return x0, y0, vals_grid 