'''
Utilities to facilitate working with PyTorch, including automatically creating
descriptive model names, saving and loading checkpoint files, and logging.
'''

import json
import logging
import os
import shutil
import subprocess as sp
import torch
import torch.nn as nn
import torch.distributed as dist
import numpy as np


class Params():
    """Class that loads hyperparameters from a json file.

    Example:
    ```
    params = Params(json_path)
    print(params.learning_rate)
    params.learning_rate = 0.5  # change the value of learning_rate in params
    ```
    """

    def __init__(self, json_path):
        with open(json_path) as f:
            params = json.load(f)
            self.__dict__.update(params)

    def save(self, json_path):
        with open(json_path, 'w') as f:
            json.dump(self.__dict__, f, indent=4)
            
    def update(self, json_path):
        """Loads parameters from json file"""
        with open(json_path) as f:
            params = json.load(f)
            self.__dict__.update(params)

    @property
    def dict(self):
        """Gives dict-like access to Params instance by `params.dict['learning_rate']"""
        return self.__dict__


class RunningAverage():
    """A simple class that maintains the running average of a quantity
    
    Example:
    ```
    loss_avg = RunningAverage()
    loss_avg.update(2)
    loss_avg.update(4)
    loss_avg() = 3
    ```
    """
    def __init__(self):
        self.steps = 0
        self.total = 0
    
    def update(self, val):
        self.total += val
        self.steps += 1
    
    def __call__(self):
        return self.total/float(self.steps)
        
    
def set_logger(log_path):
    """Set the logger to log info in terminal and file `log_path`.

    In general, it is useful to have a logger so that every output to the terminal is saved
    in a permanent file. Here we save it to `model_dir/train.log`.

    Example:
    ```
    logging.info("Starting training...")
    ```

    Args:
        log_path: (string) where to log
    """
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    if not logger.handlers:
        # Logging to a file
        file_handler = logging.FileHandler(log_path)
        file_handler.setFormatter(logging.Formatter('%(asctime)s:%(levelname)s: %(message)s'))
        logger.addHandler(file_handler)

        # Logging to console
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(logging.Formatter('%(message)s'))
        logger.addHandler(stream_handler)


def save_dict_to_json(d, json_path):
    """Saves dict of floats in json file

    Args:
        d: (dict) of float-castable values (np.float, int, float, etc.)
        json_path: (string) path to json file
    """
    with open(json_path, 'w') as f:
        # We need to convert the values to float for json (it doesn't accept np.array, np.float, )
        d = {k: float(v) for k, v in d.items()}
        json.dump(d, f, indent=4)


def save_checkpoint(state, is_best, checkpoint, extra_keys={}):
    """Saves model and training parameters at checkpoint + 'last.pth.tar'. If is_best==True, also saves
    checkpoint + 'best.pth.tar'

    Args:
        state: (dict) contains model's state_dict, may contain other keys such as epoch, optimizer state_dict
        is_best: (bool) True if it is the best model seen till now
        checkpoint: (string) folder where parameters are to be saved
        extra_keys: (dict, default empty): extra metadata about the experiment
    """
    filepath = os.path.join(checkpoint, 'last.pth.tar')
    if not os.path.exists(checkpoint):
        print("Checkpoint Directory does not exist! Making directory {}".format(checkpoint))
        os.mkdir(checkpoint)
    else:
        print("Checkpoint Directory exists! ")
    torch.save(state, filepath)
    if is_best:
        shutil.copyfile(filepath, os.path.join(checkpoint, 'best.pth.tar'))



def load_checkpoint(checkpoint, model, optimizer=None, device=None):
    """Loads model parameters (state_dict) from file_path. If optimizer is 
    provided, loads state_dict of optimizer assuming it is present in checkpoint.

    Args:
        checkpoint: (string) filename which needs to be loaded
        model: (torch.nn.Module) model for which the parameters are loaded
        optimizer: (torch.optim) optional: resume optimizer from checkpoint
    """
    if not os.path.exists(checkpoint):
        raise FileNotFoundError("File doesn't exist {}".format(checkpoint))
    if torch.cuda.is_available():
        checkpoint = torch.load(checkpoint, map_location=device)
    else:
        checkpoint = torch.load(checkpoint, map_location=torch.device('cpu'))
    
    try:
        model.load_state_dict(checkpoint['state_dict'])
    except RuntimeError:
        model = nn.DataParallel(model)
        model.load_state_dict(checkpoint['state_dict'])

    if device:
        # this helps ensure the optimizer is correctly mapped to a GPU.
        # see here: https://github.com/pytorch/pytorch/issues/8741#issuecomment-496907204
        model.to(torch.device(device))

    if optimizer:
        optimizer.load_state_dict(checkpoint['optim_dict'])

    return checkpoint


# https://nvidia.custhelp.com/app/answers/detail/a_id/3751/~/useful-nvidia-smi-queries
def get_gpu_memory():
    command = "nvidia-smi --query-gpu=memory.free --format=csv"
    free_memory = sp.check_output(command.split()).decode('ascii').split('\n')[1].split()[0]
    command = "nvidia-smi --query-gpu=memory.total --format=csv"
    total_memory = sp.check_output(command.split()).decode('ascii').split('\n')[1].split()[0]
    return {'free_mem': int(free_memory), 'total_mem': int(total_memory)}


def init_cuda(seed=1):
    c = torch.cuda.is_available()
    if c:
        print('GPU available, using it.')
        torch.cuda.manual_seed(seed)
        print(get_gpu_memory())
    else:
        print('No GPU available.')
    return c
 
def setup(rank, world_size):
    '''
    Setup for distributed data parallelism.
    See here: https://pytorch.org/tutorials/intermediate/ddp_tutorial.html
    '''
    os.environ['MASTER_ADDR'] = 'localhost'
    os.environ['MASTER_PORT'] = '12355'

    # initialize the process group
    dist.init_process_group("gloo", rank=rank, world_size=world_size)


def cleanup():
    '''
    Cleanup distributed data parallelism. 
    See here: https://pytorch.org/tutorials/intermediate/ddp_tutorial.html
    '''
    dist.destroy_process_group()


def accuracy(outputs, labels):
    """
    Compute the accuracy, given the outputs and labels for all images.

    Returns: (float) accuracy in [0,1]
    """
    outputs = np.argmax(outputs, axis=1)
    return np.sum(outputs==labels)/float(labels.size)


def make_model_name(
        m, base_dir='/home/sgreenhill/repos/WOTUS/new_imagery/experiments'
    ):
    '''
    Given a metadata dictionary, create a model name
    Args:
        m: dictionary of metadata
    '''
    # sub-directory: architecture/rule/split
    dir = f"{m['architecture']}/{m['rule']}/{m['split_slug']}"

    if m['cowardin'] or m['cowardin_simple'] or m['resource']:
        dir += '/multi_task'
    if m['ajd']:
        dir += '/ajd'

    # name: input data
    keys = list(m)
    naip_id = 12
    name = '_'.join([k for k in keys[naip_id:] if m[k]])

    if m['subset_layers']:
        name = name + '_layersubset'

    model_dir = '/'.join([base_dir, dir, name])
    print(f'Will save to: {model_dir}')
    restore_path = f'{model_dir}/best.pth.tar'

    return model_dir, restore_path