'''
Utilities for training and predicting using PyTorch deep learning models.
'''

import os
import sys
import re
import logging
import numpy as np
import pandas as pd
from tqdm import tqdm
from datetime import datetime
import torch
import torch.nn as nn
from torch.nn.functional import softmax
from torch.nn.parallel import DistributedDataParallel as DDP
import torch.distributed as dist
src_path = '/home/sgreenhill/repos/WOTUS/new_imagery/3_src'
sys.path.append(src_path)
os.chdir(src_path)
import torch_utils
import plots


def summarize(metrics, pointids_batch, loss, batch_number,
              targets=['wotus'],
              wotus_output=None, wotus_softmax=None, wotus_batch=None, 
              ajd_output=None, ajd_softmax=None, ajd_batch=None,
              cowardin_output=None, cowardin_softmax=None, cowardin_batch=None,
              resource_output=None, resource_softmax=None, resource_batch=None,
              wotus_loss=None, ajd_loss=None, cowardin_loss=None, resource_loss=None
              ):
    '''
    Calculate summary statistics on a batch
    Args:
        targets: list containing at least one of 'wotus', 'cowardin', 'resource',
            and 'ajd'
    '''
    # compute all metrics on this batch
    summary_batch = {}
    summary_batch['pointid'] = pointids_batch.tolist()

    if 'wotus' in targets:
        summary_batch['probability_wotus'] = wotus_softmax[:, 1].tolist()
        summary_batch['predictions'] = np.argmax(wotus_output, axis=1).tolist()
        if wotus_batch is not None:
            summary_batch['preds_batch'] = [batch_number for _ in range(len(wotus_output))]
            summary_batch.update({metric: metrics[metric](wotus_output, wotus_batch)
                                  for metric in metrics})
            summary_batch['loss'] = loss.item()
            summary_batch['labels'] = wotus_batch.tolist()
            summary_batch['wotus_loss'] = wotus_loss.item()
    if 'ajd' in targets:
        summary_batch['probability_ajd'] = ajd_softmax[:, 1].tolist()
        summary_batch['ajd_predictions'] = np.argmax(ajd_output, axis=1).tolist()
        if ajd_batch is not None:
            summary_batch['preds_batch'] = [batch_number for _ in range(len(ajd_output))]
            summary_batch.update({f'ajd_{metric}': metrics[metric](ajd_output, ajd_batch)
                                  for metric in metrics})
            summary_batch['loss'] = loss.item()
            summary_batch['ajd_labels'] = ajd_batch.tolist()
            summary_batch['ajd_loss'] = ajd_loss.item()
    if 'cowardin' in targets:
        summary_batch['cowardin_probabilities'] = cowardin_softmax
        summary_batch['cowardin_predictions'] = np.argmax(cowardin_output, axis=1).tolist()
        summary_batch['preds_batch'] = [batch_number for _ in range(len(cowardin_output))]
        if cowardin_batch is not None:
            summary_batch.update({metric: metrics[metric](cowardin_output, cowardin_batch)
                                  for metric in metrics})
            summary_batch['loss'] = loss.item()
            summary_batch.update({
                f'cowardin_{metric}': metrics[metric](cowardin_output, cowardin_batch)
                for metric in metrics
            })
            summary_batch['cowardin_loss'] = cowardin_loss.item()
            summary_batch['cowardin_labels'] = cowardin_batch.tolist()
    if 'resource' in targets:
        summary_batch['resource_probabilities'] = resource_softmax
        summary_batch['resource_predictions'] = np.argmax(resource_output, axis=1).tolist()
        summary_batch['preds_batch'] = [batch_number for _ in range(len(resource_output))]
        if resource_batch is not None:
            summary_batch.update({metric: metrics[metric](resource_output, resource_batch)
                                  for metric in metrics})
            summary_batch['loss'] = loss.item()
            summary_batch.update({
                f'resource_{metric}': metrics[metric](resource_output, resource_batch)
                for metric in metrics
            })
            summary_batch['resource_loss'] = resource_loss.item()
            summary_batch['resource_labels'] = resource_batch.tolist()
    
    summary_batch['metrics_batch'] = batch_number
        
    return summary_batch


def run_model(model, optimizer, loss_fn, dataloader, metrics,
              save_summary_steps, mode, targets=['wotus'],
              testing=False,
              cuda_device=0, profiler=None):
    '''
    Train a model.
    Args:
        targets: list containing at least one of 'wotus', 'cowardin', 'resource', and 'ajd'
        testing: bool. Break after two iterations?
    '''

    # Set the random seed for reproducible experiments
    torch.manual_seed(230)

    # # use GPU if available
    cuda = torch.cuda.is_available()

    if cuda:
        # torch.cuda.set_device(cuda_device)
        torch.cuda.manual_seed_all(230)

    if mode == 'train':
        model.train()
    elif mode in ['eval', 'predict']:
        model.eval()
    else:
        raise ValueError("Mode must be train, eval, or predict.")

    if mode == 'predict':
        torch.set_grad_enabled(False) # equivalent to torch.no_grad()

    # summary for current training loop and a running average object for loss
    summ = {metric: [] for metric in metrics}
    summ['loss'] = []
    summ['pointid'] = []
    summ['preds_batch'] = []
    summ['metrics_batch'] = []

    if 'wotus' in targets:
        summ['predictions'] = []
        summ['labels'] = []
        summ['probability_wotus'] = []
        summ['wotus_loss'] = []
    if 'ajd' in targets:
        summ['ajd_predictions'] = []
        summ['ajd_labels'] = []
        summ['probability_ajd'] = []
        summ['ajd_loss'] = []
        summ.update({f'ajd_{metric}': [] for metric in metrics})
    if 'cowardin' in targets:
        summ['cowardin_probabilities'] = []
        summ['cowardin_predictions'] = []
        summ['cowardin_labels'] = []
        summ['cowardin_loss'] = []
        summ.update({f'cowardin_{metric}': [] for metric in metrics})
    if 'resource' in targets:
        summ['resource_probabilities'] = []
        summ['resource_predictions'] = []
        summ['resource_labels'] = []
        summ['resource_loss'] = []
        summ.update({f'resource_{metric}': [] for metric in metrics})
        
    if mode == 'train':
        loss_avg = torch_utils.RunningAverage()

    # Use tqdm for progress bar
    with tqdm(total=len(dataloader), dynamic_ncols=True) as t:
        for i, sample_batched in enumerate(dataloader):
            data_batch = sample_batched['image']
            
            if mode != 'predict':
                if 'wotus' in targets:
                    wotus_batch = sample_batched['wotus']
                if 'ajd' in targets:
                    ajd_batch = sample_batched['ajd']
                if 'cowardin' in targets:
                    cowardin_batch = sample_batched['cowardin']
                if 'resource' in targets:
                    resource_batch = sample_batched['resource']
            else:
                if 'wotus' in targets:
                    wotus_batch = None
                if 'ajd' in targets:
                    ajd_batch = None
                if 'cowardin' in targets:
                    cowardin_batch = None
                if 'resource' in targets:
                    resource_batch = None
            
            pointids_batch = sample_batched['pointid']
            
            # move to GPU if available
            if cuda:
                data_batch = data_batch.cuda(device=cuda_device, non_blocking=True)
                if mode != 'predict':
                    if 'wotus' in targets:
                        wotus_batch = wotus_batch.cuda(device=cuda_device, non_blocking=True)
                    if 'ajd' in targets:
                        ajd_batch = ajd_batch.cuda(device=cuda_device, non_blocking=True)
                    if 'cowardin' in targets:
                        cowardin_batch = cowardin_batch.cuda(device=cuda_device, non_blocking=True)
                    if 'resource' in targets:
                        resource_batch = resource_batch.cuda(device=cuda_device, non_blocking=True)
                
                model = model.cuda(device=cuda_device)

            # compute model output and loss
            if len(targets) == 1:
                if 'wotus' in targets:
                    wotus_output = model(data_batch)
                elif 'ajd' in targets:
                    ajd_output = model(data_batch)
                elif 'cowardin' in targets:
                    cowardin_output = model(data_batch)
                elif 'resource' in targets:
                    resource_output = model(data_batch)
                else:
                    raise ValueError('Unrecognized target.')
            else:
                raise ValueError('Only implemented for 1-target. See git history for multi-target infrastructure')

            if mode != 'predict':
                if 'wotus' in targets:
                    wotus_batch = wotus_batch.squeeze(axis=1)
                    wotus_loss = loss_fn(wotus_output, wotus_batch)
                    if len(targets) == 1:
                        loss = wotus_loss
                if 'ajd' in targets:
                    ajd_batch = ajd_batch.squeeze(axis=1)
                    ajd_loss = loss_fn(ajd_output, ajd_batch)
                    if len(targets) == 1:
                        loss = ajd_loss
                        if np.isnan(float(loss)):
                            from pdb import set_trace; set_trace()
                if 'cowardin' in targets:
                    cowardin_batch = cowardin_batch.squeeze(axis=1).type(torch.long)
                    cowardin_loss = loss_fn(cowardin_output, cowardin_batch)
                    if len(targets) == 1:
                        loss = cowardin_loss
                if 'resource' in targets:
                    resource_batch = resource_batch.squeeze(axis=1).type(torch.long)
                    resource_loss = loss_fn(resource_output, resource_batch)
                    if len(targets) == 1:
                        loss = resource_loss
                elif len(targets) != 1:
                    raise ValueError('Loss not defined. Only implemented for 1-target.')
            else:
                loss = np.nan
                wotus_loss = np.nan
                cowardin_loss = np.nan
                resource_loss = np.nan
            
            if 'wotus' in targets:
                wotus_softmax = softmax(wotus_output, dim=1)
            if 'ajd' in targets:
                ajd_softmax = softmax(ajd_output, dim=1)
            if 'cowardin' in targets:
                cowardin_softmax = softmax(cowardin_output, dim=1)
            if 'resource' in targets:
                resource_softmax = softmax(resource_output, dim=1)

            if mode == 'train':
                # clear previous gradients, compute gradients of all variables wrt loss
                optimizer.zero_grad()
                loss.backward()

                # performs updates using calculated gradients
                optimizer.step()

            # Evaluate summaries only once in a while
            if i % save_summary_steps == 0 or mode == 'eval':
                # extract data from torch Variable, move to cpu, convert to numpy arrays
                if cuda:
                    if 'wotus' in targets:
                        wotus_output = wotus_output.data.cpu().numpy()
                        wotus_softmax = wotus_softmax.data.cpu().numpy()
                    if 'ajd' in targets:
                        ajd_output = ajd_output.data.cpu().numpy()
                        ajd_softmax = ajd_softmax.data.cpu().numpy()
                    if 'cowardin' in targets:
                        cowardin_output = cowardin_output.data.cpu().numpy()
                        cowardin_softmax = cowardin_softmax.data.cpu().numpy()
                    if 'resource' in targets:
                        resource_output = resource_output.data.cpu().numpy()
                        resource_softmax = resource_softmax.data.cpu().numpy()
                    if mode != 'predict':
                        if 'wotus' in targets:
                            wotus_batch = wotus_batch.data.cpu().numpy()
                        if 'ajd' in targets:
                            ajd_batch = ajd_batch.data.cpu().numpy()
                        if 'cowardin' in targets:
                            cowardin_batch = cowardin_batch.data.cpu().numpy()
                        if 'resource' in targets:
                            resource_batch = resource_batch.data.cpu().numpy()

                else:
                    if 'wotus' in targets:
                        wotus_output = wotus_output.data.numpy()
                        wotus_softmax = wotus_softmax.data.numpy()
                    if 'ajd' in targets:
                        ajd_output = ajd_output.data.numpy()
                        ajd_softmax = ajd_softmax.data.numpy()
                    if 'cowardin' in targets:
                        cowardin_output = cowardin_output.data.numpy()
                        cowardin_softmax = cowardin_softmax.data.numpy()
                    if 'resource' in targets:
                        resource_output = resource_output.data.numpy()
                        resource_softmax = resource_softmax.data.numpy()
                    if mode != 'predict':
                        if 'wotus' in targets:
                            wotus_batch = wotus_batch.data.numpy()
                        if 'ajd' in targets:
                            ajd_batch = ajd_batch.data.numpy()
                        if 'cowardin' in targets:
                            cowardin_batch = cowardin_batch.data.numpy()
                        if 'resource' in targets:
                            resource_batch = resource_batch.data.numpy()

                # compute all metrics on this batch
                if len(targets) == 1:
                    if 'wotus' in targets:
                        summary_batch = summarize(
                            metrics, pointids_batch, loss, i, targets=targets,
                            wotus_loss=wotus_loss, wotus_output=wotus_output,
                            wotus_softmax=wotus_softmax, wotus_batch=wotus_batch
                        )
                    elif 'ajd' in targets:
                        summary_batch = summarize(
                            metrics, pointids_batch, loss, i, targets=targets,
                            ajd_loss=ajd_loss, ajd_output=ajd_output,
                            ajd_softmax=ajd_softmax, ajd_batch=ajd_batch
                        )
                    elif 'cowardin' in targets:
                        summary_batch = summarize(
                            metrics, pointids_batch, loss, i, targets=targets,
                            cowardin_output=cowardin_output, 
                            cowardin_softmax=cowardin_softmax,
                            cowardin_batch=cowardin_batch,
                            cowardin_loss=cowardin_loss,
                        )
                    elif 'resource' in targets:
                        summary_batch = summarize(
                        metrics, pointids_batch, loss, i, targets=targets,
                        resource_output=resource_output,
                        resource_softmax=resource_softmax,
                        resource_batch=resource_batch,
                        resource_loss=resource_loss                        
                    )
                    else:
                        raise ValueError('Target not recognized')
                else:
                    raise ValueError('Only implemented for 1-target.')

                for k, v in summary_batch.items():
                    if not isinstance(v, list):
                        v = [v]
                    summ[k] += v

            if mode == 'train':
                # update the average loss
                loss_avg.update(loss.item())
                t.set_postfix(loss='{:05.3f}'.format(loss_avg()))
            t.update()

            if i == 2 and testing:
                print('Breaking for testing')
                break # FOR TESTING

            if profiler is not None:
                profiler.step()
                if i > 10:
                    print('Breaking for profiler')
                    break

    torch.cuda.empty_cache()
    return summ



def train_and_evaluate(model, dataloaders, optimizer, loss_fn, 
                       num_epochs, metrics, model_dir, restore_path=None, 
                       save_summary_steps=1,
                       targets=['wotus'],
                       testing=False, cuda_device=0,
                       checkpoint_metadata={}, parallelize_gpus=True,
                       profiler=None, scheduler=None
                       ):
    '''
    Train a model and evaluate every epoch.
    Args:
        model: (torch.nn.Module) the neural network
        dataloaders: (dict) a dictionary with keys "train" and "val" and values 
         of a torch.utils.data.DataLoader object that fetch data
        optimizer: (torch.optim) optimizer for parameters of model
        loss_fn: a function that takes batch_output and batch_labels and 
            computes the loss for the batch
        metrics: (dict) a dictionary of functions that compute a metric using 
            the output and labels of each batch
        model_dir: (string) directory containing config, weights and log
        restore_path: (string) optional- name of file to restore from 
            (without its extension .pth.tar)
        targets: list containing at least one of 'wotus', 'cowardin', and 'resource'       
    '''

    # set the default cuda device
    cuda = torch.cuda.is_available()
    if cuda:
        gpus = torch.cuda.device_count()
        if gpus > 1 and parallelize_gpus:
            if isinstance(cuda_device, list):
                  print(f'Parallelizing over {len(cuda_device)} GPUs, using ids {cuda_device}') 
                  default_cuda_device = cuda_device[0]
                  print(f'Default device is {default_cuda_device}.') 
            else:
                print(f'Parallelizing over {gpus} GPUs')
                default_cuda_device=0
        else:
            print(f'Using GPU {cuda_device}')
            torch.cuda.set_device(cuda_device)
            default_cuda_device=0
    else:
        print('Not using GPU')
        default_cuda_device = False

    # reload weights from restore_file if specified
    if restore_path is not None:
        try:
            checkpoint = torch_utils.load_checkpoint(
                restore_path, model, optimizer=optimizer,
                device=f'cuda:{default_cuda_device}'
            )
            last_epoch = checkpoint['epoch']
            logging.info("Restored parameters from {}".format(restore_path))
        except FileNotFoundError:
            logging.info(f"{restore_path} does not exist. Starting training from scratch.")
            last_epoch = 0
    else:
        last_epoch = 0

    best_val_loss = 1e4

    if cuda:
        gpus = torch.cuda.device_count()
        if gpus > 1 and parallelize_gpus:
            if isinstance(cuda_device, list):
                model = nn.DataParallel(model, device_ids=cuda_device)
            else:
                model = nn.DataParallel(model)

    for epoch in range(last_epoch, num_epochs):
        print(f'Epoch {epoch}')
        # Run one epoch
        logging.info("Epoch {}/{}".format(epoch + 1, num_epochs))

        # compute number of batches in one epoch (one full pass over the training set)
        print('Training...')
        print(f'Testing is {testing}')
        train_metrics = run_model(
            model, optimizer, loss_fn, dataloaders['train'], metrics, 
            save_summary_steps, mode='train',
            targets=targets, testing=testing, 
            cuda_device=default_cuda_device,
            profiler=profiler
        )

        if not os.path.exists(model_dir):
            os.makedirs(model_dir)

        # Evaluate for one epoch on validation set
        print('Evaluating...')
        val_metrics = run_model(
            model, optimizer, loss_fn, dataloaders['val'], metrics,
            save_summary_steps, mode='eval', testing=testing,
            targets=targets, cuda_device=default_cuda_device
        )
        
        val_loss = np.mean(val_metrics['loss'])
        print(f'Average val loss this epoch: {val_loss}')
        
        if scheduler is not None:
            scheduler.step(val_loss)
        
        if 'wotus' in targets:
            print(f'WOTUS accuracy this epoch: {np.mean(val_metrics["accuracy"])}')
        if 'ajd' in targets:
            print(f'AJD accuracy this epoch: {np.mean(val_metrics["ajd_accuracy"])}')
        if 'cowardin' in targets:
            print(f'Cowardin accuracy this epoch: {np.mean(val_metrics["cowardin_accuracy"])}')
        if 'resource' in targets:
            print(f'Resource type accuracy this epoch: {np.mean(val_metrics["resource_accuracy"])}')

        is_best = val_loss < best_val_loss

        # save batch-wise results
        fold_names = ['train', 'val']
        res_names = ['preds', 'metrics']
        for i, res in enumerate([train_metrics, val_metrics]):
            metrics_keys = ['loss']
            res_keys = ['pointid']
            if 'wotus' in targets:
                metrics_keys += ['accuracy', 'loss', 'metrics_batch', 'wotus_loss']
                res_keys += [
                    'predictions', 'labels', 'preds_batch', 'probability_wotus'
                ]
            if 'ajd' in targets:
                metrics_keys += ['ajd_accuracy', 'ajd_loss']
                res_keys += ['ajd_predictions', 'probability_ajd', 'ajd_labels']
            if 'cowardin' in targets:
                metrics_keys += ['cowardin_loss', 'cowardin_accuracy']
                res_keys += [
                    'cowardin_predictions', 'cowardin_probabilities',
                    'cowardin_labels'
                ]
            if 'resource' in targets:
                metrics_keys += ['resource_loss', 'resource_accuracy']
                res_keys += [
                    'resource_predictions', 'resource_probabilities',
                    'resource_labels'
                ]

            res_metrics = {k: res[k] for k in metrics_keys}
            res_preds = {k: res[k] for k in res_keys}
            for j, df in enumerate([res_preds, res_metrics]):
                # do some special handling for the cowardin and resource probabilities,
                # since these need many columns for the different classes
                if 'cowardin' in targets and j == 0:
                    cowardin_p_df = pd.DataFrame(
                        np.concatenate(
                            df.pop('cowardin_probabilities')
                        )
                    ).add_prefix('cowardin_probability_')
                if 'resource' in targets and j == 0:
                    resource_p_df = pd.DataFrame(
                        np.concatenate(
                            df.pop('resource_probabilities')
                        )
                    ).add_prefix('resource_probability_')                  

                df = pd.DataFrame(df)

                if 'cowardin' in targets and j == 0:
                    df = pd.concat([df, cowardin_p_df], axis=1)
                if 'resource' in targets and j == 0:
                    df = pd.concat([df, resource_p_df], axis=1)

                df['epoch'] = epoch
                fname = f'{model_dir}/{fold_names[i]}_{res_names[j]}.csv'
                if epoch == 0:
                    df.to_csv(fname, index=False)
                else:
                    df.to_csv(fname, index=False, header=False, mode='a')

        # If best_eval, best_save_path
        if is_best and not testing:
            print('This is the best average val loss so far.')
            best_val_loss = val_loss
            
            # Save weights
            checkpoint_dict = {
                'epoch': epoch + 1,
                'state_dict': model.state_dict(),
                'optim_dict': optimizer.state_dict(),
                'date_run': datetime.now()
            }
            checkpoint_dict.update(checkpoint_metadata)

            torch_utils.save_checkpoint(checkpoint_dict, is_best=is_best,
                                        checkpoint=model_dir)

    # Lastly, print and plot results
    train_metrics = pd.read_csv(f'{model_dir}/train_metrics.csv')
    val_metrics = pd.read_csv(f'{model_dir}/val_metrics.csv')
    train_preds = pd.read_csv(f'{model_dir}/train_preds.csv')
    val_preds = pd.read_csv(f'{model_dir}/val_preds.csv')

    best_train = train_metrics.groupby('epoch').mean().accuracy.max()
    best_val = val_metrics.groupby('epoch').mean().accuracy.max()
    print(f'Best train accuracy: {best_train :0.2f}')
    print(f'Best val accuracy: {best_val :0.2f}')

    plots.plot_metrics(train_metrics, val_metrics, train_preds, val_preds,
                       model_dir=model_dir, title=re.split('/', model_dir)[-2])
