'''
Model with NAIP + 150k prediction points, trained on all main input layers except for
rule, pooling across rules and districts. Target: AJD
'''

import torch
import torch.nn as nn
import torch.optim as optim
from torch.optim import lr_scheduler
import torch.backends.cudnn as cudnn
import os

cudnn.benchmark = True


# my additions
from tqdm import tqdm
import numpy as np
import pandas as pd
from torch.utils.data import DataLoader
from torch.multiprocessing import cpu_count 

import sys
src_path = '/home/sgreenhill/repos/WOTUS/new_imagery/src'
sys.path.append(src_path)
os.chdir(src_path)
import torch_utils as utils
import train
import data
import models

# Configuration
num_cores = int(cpu_count())
batch_size = 16

# use GPU if available
cuda = torch.cuda.is_available()

# # Set the random seed for reproducible experiments
torch.manual_seed(230)
if cuda:
    torch.cuda.manual_seed(230)
    gpu_info = utils.get_gpu_memory()
    print(gpu_info)

path = '/home/sgreenhill/wotus'

split_path = f'{path}/misc/ajd_train_test_split.csv'
split = pd.read_csv(split_path, low_memory=False)
split_slug = 'naip_ajd'

# drop pointids where data is missing
split = data.drop_missings(split, drop_S2=False, 
                           dir=f'{path}/misc')

# drop pred points where NHD is corrupted
missing_new_nhd = np.load(f'{path}/prediction_points/new_nhd_corrupted_points.p.npy', allow_pickle=True)
# from pdb import set_trace; set_trace()
split['pred_id'] = np.where(split.id.str.startswith('pred'), split.id.str.slice(5).astype(int), np.nan)
split = split.loc[~split.pred_id.isin(missing_new_nhd.astype(int))]

# prep the data separately, then combine 
ajd_split = split.loc[split.ajd == 1]
pred_split = split.loc[split.ajd == 0]

#############
# Prep data #
#############
ajd_names, _, _, _, ajd_labels = data.make_split_dicts(
    ajd_split, wotus=False, ajd=True, verbose=False
)

folds = ['train', 'val']

normalize_path='/home/sgreenhill/repos/WOTUS/new_imagery/analysis/wotus/main_model/train/layer_mean_sd'

ajd_dataset = {x: data.NAIPDataset(
    root_dir='/home/sgreenhill/wotus/satellite_imagery',
    normalize=True, normalize_path=normalize_path,
    image_names=ajd_names[x], fold=x, ajd_labels=ajd_labels[x], wotus=False, ajd=True,
    n_naip_bands=4, naip=True, image_dir='AJD_inputs/NAIP', nwi=True, 
    nhd=True, nhd_dir='../prediction_points/NHD_new', 
    dem=True, ecoregion=True, 
    nlcd=True, prism=True, gnatsgo=True, hq_dist=True, district_dummies=True, 
    regrid=False, rule_dummies=False, states=True,
    save=False, load_saved=True,
    save_dir=f'/home/sgreenhill/wotus/satellite_imagery/pipeline_ready/ajd_new_nhd'
) for x in folds}

# need to loop over all 50 grids here
pred_datasets = []
for grid in range(1, 51):
    this_split = pred_split.loc[pred_split.prcss_r == grid]
    
    if this_split.shape[0] == 0:
        continue

    pred_names, _, _, _, pred_labels =  data.make_split_dicts(
        this_split, wotus=False, ajd=True, verbose=False
    )

    pred_dir = f'/home/sgreenhill/wotus/prediction_points/'

    pred_datasets += [{x: data.NAIPDataset(
        image_names=pred_names[x], fold=x, ajd_labels=pred_labels[x],
        normalize=True, normalize_path=normalize_path,
        wotus=False, ajd=True, n_naip_bands=4, 
        naip=True, nwi=True, nhd=True, dem=True, ecoregion=True, nlcd=True, prism=True,
        gnatsgo=True, hq_dist=True, 
        district_dummies=True, regrid=False, rule_dummies=False, states=True,
        root_dir= f'{pred_dir}/grid_{grid}',
        image_dir = f'NAIP/', image_prefix='NAIP_',
        nwi_dir='NWI', nhd_dir='../NHD_new_prediction/', dem_dir='DEM', 
        ecoregion_dir = 'Ecoregion', nlcd_dir='NLCD', prism_dir='PRISM', 
        prism_ppt_dir='PPT', gnatsgo_dir='gNATSGO', 
        dem_prefix='DEM_', ecoregion_prefix='Ecoregion_',
        gnatsgo_prefix='gNATSGO_',
        nlcd_prefix='NLCD_', prism_prefix='PRISM_', 
        prism_ppt_prefix='PPT_', gnatsgo_stacked=True,
        hq_dist_dir='../HQ_dist/', hq_dist_prefix='',
        rule_dummies_dir=None, rule_dummies_prefix='',
        states_dir='../state', states_prefix='', 
        district_dummies_dir='../ACE_districts', district_dummies_prefix='',
        save=True, load_saved=True,
        save_dir=f'/home/sgreenhill/wotus/satellite_imagery/pipeline_ready/pred/grid_{grid}'
    ) for x in folds}]

all_datasets = [ajd_dataset] + pred_datasets

dataloaders = {
    x: torch.utils.data.DataLoader(
        data.ConcatDataset(
            [d[x] for d in all_datasets]
        ),
        batch_size=batch_size,
        shuffle=True, 
        num_workers=num_cores 
    ) for x in folds
}

###############
# Train model #
###############

# hyperparameters
learning_rate = 1e-3
regularization = 1e-4
num_epochs = 30

model = models.resnet18(in_channels=31, n_classes=2)

optimizer = optim.Adam(model.parameters(), lr=learning_rate,
                    weight_decay=regularization)

scheduler = optim.lr_scheduler.ReduceLROnPlateau(
    optimizer, patience=3, threshold=0.001, threshold_mode='abs',
    verbose=True, factor=0.5
)

loss_fn = nn.CrossEntropyLoss()
metrics = {'accuracy': utils.accuracy}

config = {
    'architecture': 'resnet-18',
    'rule': 'pooled',
    'split': split_path,
    'split_slug': split_slug,
    'learning_rate': learning_rate,
    'weight_decay': regularization,
    'num_epochs': num_epochs
}

data_config = {
    'batch_size': batch_size, 'augmented': True, 'normalized': True,
    'shuffled': True, 'subset_layers': False,
    'naip4': 'naip4',
    'nwi': True, 'nhd': True, 'dem': True, 'ecoregion': True, 
    'nlcd': True, 'prism': True, 'gnatsgo': True, 
    'district_dummies': True, 'hq_dist': True,
    'regrid': False, 'rule_dummies': False, 'states': True,
    'cowardin': False, 'cowardin_simple': False, 
    'resource': False, 'ajd': True
}

config.update(data_config)

model_dir = '/home/sgreenhill/repos/WOTUS/new_imagery/analysis/ajd/model_results'
restore_path = f'{model_dir}/best.pth.tar'

# print(model_dir)

save_summary_steps = 1

train.train_and_evaluate(
    model, dataloaders, optimizer, loss_fn, num_epochs, metrics, model_dir,
    restore_path, save_summary_steps, checkpoint_metadata=config,
    scheduler=scheduler, testing=False,
    parallelize_gpus=True, cuda_device=[0, 1, 2],
    targets=['ajd']
)
