'''
Predict on grid points, using data that has been moved to an SSD for faster IO.
'''

import sys
import os
src_path = '/home/sgreenhill/repos/WOTUS/new_imagery/src'
sys.path.append(src_path)
os.chdir(src_path)
import torch_utils
import models
import data
import train
import torch
import torch.nn as nn
import torch.optim as optim
from torch.multiprocessing import cpu_count
import numpy as np
import pandas as pd
from tqdm import tqdm
import multiprocessing as mp
from multiprocessing.pool import ThreadPool
import shutil
from functools import partial
from glob import glob


class MultithreadedCopier:
    '''
    Multi-threaded copying for fast moving of data
    '''
    def __init__(self, max_threads):
        self.pool = ThreadPool(max_threads)

    def copy(self, source, dest):
        self.pool.apply_async(shutil.copy2, args=(source, dest))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.pool.close()
        self.pool.join()

model_dir = '/ssd-mount/'
model_name = 'model_results/'

chkpt_file = model_dir + model_name + 'best.pth.tar'
model = models.resnet18(in_channels=34, n_classes=2)
learning_rate = 1e-3
regularization = 1e-4

default_cuda_device = 'cuda:0'
chkpt = torch_utils.load_checkpoint(chkpt_file, model, device=torch.device(default_cuda_device))

model = nn.DataParallel(model, device_ids=[0, 1, 2, 3])

model.eval()

batch_size = 1
num_cores =  int(cpu_count())

# load the data we want to predict on
pred_dir = f'/ssd-mount/'
pred_pts = pd.read_csv(f'{pred_dir}/sample_grid_points.csv')

for grid in range(1, 51):
    print(f'----GRID {grid}----')
    rules = []
    for rule in ['Rapanos', 'NWPR', 'CWR']:
        if os.path.exists(f'{model_dir}/{model_name}/grid{grid}_predictions_{rule}.csv'):
            print(f'Output file already exists for grid {grid} {rule}. Skipping.')
        else:
            rules += [rule]
    if len(rules) == 0:
        print(f'All output files already exist for grid {grid}. Moving on.')
        continue
    
    print('---Moving data to SSD---')
    try:
        print('Grid points...')
        # copy the points that are organized by grid
        if not os.path.exists(f'/ssd-mount/grid_{grid}'):
            with MultithreadedCopier(max_threads=num_cores) as copier:
                shutil.copytree(
                    f'/home/sgreenhill/wotus/prediction_points/grid_points/grid_{grid}/',
                    f'/ssd-mount/grid_{grid}/',
                    copy_function=copier.copy
                )

        pred_pts = pd.read_csv(f'{pred_dir}/sample_grid_points.csv')
        missings = np.load(f'{pred_dir}/missing_fe_pts.npy')
        missings_nhd = np.load(f'{pred_dir}/new_nhd_corrupted_points.p.npy', allow_pickle=True)
        these_pred_pts = pred_pts.loc[
            (pred_pts.process_order == grid) & 
            (~pred_pts.prediction_id.astype(str).isin(missings)) &
            (~pred_pts.prediction_id.astype(str).isin(missings_nhd))
            ]

        # copy only the desired points from each of the folders where all the data are
        # saved in one folder
        old_pred_dir = f'/home/sgreenhill/wotus/prediction_points/grid_points'
        for folder in ['NHD_new_prediction', 'ACE_districts', 'HQ_dist', 'state',
                       'rules/Rapanos', 'rules/NWPR', 'rules/CWR']:
            print(f'{folder}')

            # get files, accounting for points already present
            if folder == 'NHD_new_prediction':
                files = [f'/home/hdruckenmiller/{folder}/NHD_{str(p)}.tiff' 
                         for p in these_pred_pts.prediction_id]
            else:
                files = [f'{old_pred_dir}/{folder}/{str(p)}.tif' 
                         for p in these_pred_pts.prediction_id]

            copier = partial(
                shutil.copy2,
                dst=f'/ssd-mount/{folder}'
            )

            with mp.Pool(processes=num_cores) as p:
                l = list(tqdm(p.imap(copier, files), total=len(files)))
    except Exception as e:
        print(f'Failed to load data for grid {grid}. Skipping.')
        print(f'Error was {e}.')
    try:
        print('---Predicting---')
        for rule in rules:
            print(f'Starting {rule}')
            
            # filter to the grid 
            these_pred_pts = pred_pts.loc[pred_pts.process_order == grid]

            ids = these_pred_pts.prediction_id.astype(str)

            # drop missings
            # files entirely missing from main layers
            missings = np.load(f'{pred_dir}/grid_{grid}/missing_pts.npy', allow_pickle=True)

            # corrupted files from the new NHD data
            missing_new_nhd = np.load(f'{pred_dir}/corrupted_new_nhd_points.npy', allow_pickle=True)
            # files entirely missing from HQ dist, rule, state, district FE layers
            fe_missings = np.load(f'{pred_dir}/missing_fe_pts.npy', allow_pickle=True)
            # corrupted rule, HQ dist, state, ace districts files
            rule_missings = np.load(f'{pred_dir}/rules/rule_drops_{rule}.npy', allow_pickle=True)
            districts_missings = np.load(f'{pred_dir}/corrupted_ace_district_points.npy', allow_pickle=True)
            hq_dist_missings = np.load(f'{pred_dir}/corrupted_hq_dist_points.npy', allow_pickle=True)
            state_missings = np.load(f'{pred_dir}/corrupted_state_points.npy', allow_pickle=True)

            # naip images that are the wrong shape (3 band instead of 4)
            naip_missings = np.load(f'{pred_dir}/grid_{grid}/naip_drops.npy', allow_pickle=True)

            missings = np.concatenate(
                [missings, 
                    missing_new_nhd, 
                    fe_missings, rule_missings, districts_missings, 
                    hq_dist_missings, state_missings, naip_missings]
            )

            n_predrop = len(ids)
            ids = ids[~ids.isin(missings)]
            n_postdrop = len(ids)
            print(f'Dropped {n_predrop - n_postdrop} missing ids.')

            # converting back to integer identifiers plays nicer with data.py
            ids = ids.astype(int)

            names = {'predict': list(ids)}
            labels = {'predict': None}
            cowardin_labels = {'predict': None}
            resource_labels = {'predict': None}
            ajd_labels = {'predict': None}


            datasets, dataloaders = data.get_NAIP_dataloaders(
                names, labels, batch_size, num_cores, 
                normalize=True, 
                normalize_path=f'{pred_dir}/layer_mean_sd/',
                n_naip_bands=4, naip=True,
                nwi=True, nhd=True, dem=True, ecoregion=True, nlcd=True, prism=True,
                gnatsgo=True, hq_dist=True, 
                district_dummies=True, regrid=False, rule_dummies=True, 
                rule_one_hot=True, states=True,
                return_datasets=True,
                root_dir= f'{pred_dir}/grid_{grid}',
                image_dir = f'NAIP/', image_prefix='NAIP_',
                nwi_dir='NWI/', nwi_prefix='NWI_',
                nhd_dir='../NHD_new_prediction',
                nhd_prefix='NHD_', nhd_stacked=True, dem_dir='DEM', 
                ecoregion_dir = 'Ecoregion', nlcd_dir='NLCD', prism_dir='PRISM', 
                prism_ppt_dir='PPT', gnatsgo_dir='gNATSGO', 
                dem_prefix='DEM_', ecoregion_prefix='Ecoregion_',
                gnatsgo_prefix='gNATSGO_', gnatsgo_stacked=True,
                nlcd_prefix='NLCD_', prism_prefix='PRISM_', 
                prism_ppt_prefix='PPT_', 
                hq_dist_dir='../HQ_dist/', hq_dist_prefix='',
                rule_dummies_dir=f'../rules/{rule}/', rule_dummies_prefix='',
                states_dir='../state', states_prefix='', 
                district_dummies_dir='../ACE_districts', district_dummies_prefix='',
                folds=['predict'], cowardin_labels=cowardin_labels, 
                resource_labels=resource_labels,
                ajd_labels=ajd_labels,
                shuffle=False
            )

            optimizer = optim.Adam(model.parameters(), lr=learning_rate, 
                                weight_decay=regularization)
            loss_fn = nn.CrossEntropyLoss()
            metrics = {'accuracy': torch_utils.accuracy}

            preds = train.run_model(model, optimizer, loss_fn, dataloaders['predict'],
                                    metrics, save_summary_steps=1, mode='predict', testing=False,
                                    cuda_device=default_cuda_device)

            preds_df = pd.DataFrame({k: preds[k] for k in ['pointid', 'probability_wotus']})

            # unlist the pointids
            preds_df['pointid'] = [v[0] for v in preds_df.pointid.values]

            preds_df.to_csv(f'{model_dir}/{model_name}/grid{grid}_predictions_{rule}.csv', 
                            index=False)
    except Exception as e:
        print(f'Prediction failed for grid {grid} {rule}. Error was {e}.')   
    try:
        print(f'---Removing data for grid {grid}---')
        # the grid organized data
        print('Removing grid folder...')
        shutil.rmtree(f'/ssd-mount/grid_{grid}')

        # the folders
        for folder in ['ACE_districts', 'HQ_dist', 'state',
                    'rules/Rapanos', 'rules/NWPR', 'rules/CWR', 
                    'NHD_new_prediction']:
            print(f'Emptying {folder}...')
            files = glob(f'/ssd-mount/{folder}/*.tif*')
            with mp.Pool(processes=num_cores) as p:
                l = list(tqdm(p.imap(os.remove, files), total=len(files)))

        print('Done! Moving on to next grid.')
    except Exception as e:
        print(f'Errored removing data for grid {grid}. Exception was {e}.')
