'''
Train a model that has all the inputs from GEE
'''

from __future__ import print_function, division

import torch
import torch.nn as nn
import torch.optim as optim
from torch.optim import lr_scheduler
import torch.backends.cudnn as cudnn
import matplotlib.pyplot as plt
import os

cudnn.benchmark = True
plt.ion()   # interactive mode


# my additions
from tqdm import tqdm
import numpy as np
import pandas as pd
from torch.utils.data import DataLoader
from torch.multiprocessing import cpu_count 

import sys
src_path = '/home/sgreenhill/repos/WOTUS/new_imagery/src'
sys.path.append(src_path)
os.chdir(src_path)
import torch_utils as utils
import train
import data
import models

# Configuration
num_cores = 86
batch_size = 32

# hyperparameters
learning_rate = 1e-3
regularization = 1e-4
num_epochs = 50


# use GPU if available
cuda = torch.cuda.is_available()

# # Set the random seed for reproducible experiments
torch.manual_seed(230)
if cuda:
    torch.cuda.manual_seed(230)
    gpu_info = utils.get_gpu_memory()
    print(gpu_info)

path = '/home/sgreenhill/wotus'
split_path = f'{path}/misc/train_test_split_naip_only.csv'
split = pd.read_csv(split_path)
split_slug = 'naip_simon'

split = split.loc[~split.rule.isnull()]

# drop pointids where data is missing
split = data.drop_missings(split, drop_S2=False)

#############
# Prep data #
#############

names, labels, cowardin_labels, resource_labels, ajd_labels = data.make_split_dicts(split)

datasets, dataloaders, data_config = data.get_NAIP_dataloaders(
    names, labels, batch_size, num_cores, 
    normalize=True,
    normalize_path='/home/sgreenhill/repos/WOTUS/new_imagery/4_dl_models/wotus/train/layer_mean_sd',
    n_naip_bands=4, 
    naip=True, nwi=True, 
    nhd=True, nhd_dir='../../../hdruckenmiller/NHD_new/', nhd_prefix='NHD_',
    dem=True, ecoregion=True, nlcd=True, prism=True,
    gnatsgo=True, hq_dist=True,
    district_dummies=True, regrid=False, rule_dummies=True, rule_one_hot=True,
    states=True,
    return_datasets=True, return_config=True, 
    cowardin_labels=cowardin_labels, resource_labels=resource_labels,
    ajd_labels=ajd_labels,
    save=True,
    save_dir=f'{path}/satellite_imagery/pipeline_ready/ajd_new_nhd',
    load_saved=True
)

###############
# Train model #
###############

model = models.resnet18(in_channels=34, n_classes=2)

optimizer = optim.Adam(model.parameters(), lr=learning_rate,
                    weight_decay=regularization)

scheduler = optim.lr_scheduler.ReduceLROnPlateau(
    optimizer, patience=3, threshold=0.001, threshold_mode='abs',
    verbose=True, factor=0.5
)

loss_fn = nn.CrossEntropyLoss()
metrics = {'accuracy': utils.accuracy}

config = {
    'architecture': 'resnet-18',
    'rule': 'pooled',
    'split': split_path,
    'split_slug': split_slug,
    'learning_rate': learning_rate,
    'weight_decay': regularization,
    'num_epochs': num_epochs
}

config.update(data_config)

model_dir = '/home/sgreenhill/repos/WOTUS/new_imagery/analysis/wotus/diagnostics/updated_nhd/model_results'
restore_path = f'{model_dir}/best.pth.tar'

print(model_dir)

save_summary_steps = 1

train.train_and_evaluate(
    model, dataloaders, optimizer, loss_fn, num_epochs, metrics, model_dir,
    restore_path, save_summary_steps, checkpoint_metadata=config,
    scheduler=scheduler, testing=False,
    parallelize_gpus=True
)