import numpy as np
import pandas as pd
import pickle

import torch
import torch.nn as nn
import torch.optim as optim

import sys
import os
src_path = '/home/sgreenhill/repos/WOTUS/new_imagery/3_src/'
sys.path.append(src_path)

import torch_utils
import models
import data
import train

# run config
num_cores = os.cpu_count()
batch_size = 32
learning_rate = 1e-3
regularization = 1e-4
batch_size = 1
default_cuda_device = 0

# use GPU if available
cuda = torch.cuda.is_available()

# Set the random seed for reproducible experiments
torch.manual_seed(230)
if cuda:
    torch.cuda.set_device(default_cuda_device)
    torch.cuda.manual_seed(230)
    gpu_info = torch_utils.get_gpu_memory()
    print(gpu_info)

path = '/home/sgreenhill/wotus'
split_path = f'{path}/misc/train_test_split_naip_only.csv'
split = pd.read_csv(split_path)
split_slug = 'naip_simon'

split = split.loc[~split.rule.isnull()]

# drop pointids where data is missing
split = data.drop_missings(split, drop_S2=False)

# merge in state info
jds = pd.read_csv(f'{path}/AJD/jds_wet_dry_season_clean_v2.csv')[['pointid', 'state']]
split = split.merge(jds, on='pointid', how='inner')

# load the model
chkpt_file = '/home/sgreenhill/repos/WOTUS/new_imagery/4_dl_models/wotus/model_results/best.pth.tar'
model = models.resnet18(in_channels=34, n_classes=2)
chkpt = torch_utils.load_checkpoint(chkpt_file, model, device=torch.device('cuda:{}'.format(default_cuda_device)))
model = model.eval()

########################################
# Permutation by block, national model #
########################################

# use data dict to identify which layers to permute
load_path = '/ssd-mount/ajd_new_nhd_test_set'
data_dict = np.load(f'{load_path}/data_dict.p', allow_pickle=True)

# combine nwi and nhd
nwi = data_dict.pop('nwi_id')
nhd = data_dict.pop('nhd_ids')
data_dict['nwi_nhd_ids'] = nwi + nhd

# remove augment and normalize, which are not layers and thus can't be permuted
del data_dict['augment'], data_dict['normalize']

states = np.sort(split.state.unique())
permute_accs_by_state = {}
test_obs_by_state = {}

for state in states:
    print('#'*(len(state) + 4))
    print(f'# {state} #')
    print('#'*(len(state) + 4))
    
    split_state = split[split.state == state].reset_index(drop=True)
    test_obs = split_state.loc[split_state.split_group == 'test'].shape[0]
    if test_obs == 0:
        print('No obs. Skipping')
        continue

    test_obs_by_state[state] = test_obs
    
    state_permute_accs_blocks = {}

    for key in data_dict:
        permute_layer = data_dict[key]
        print(f'----Permuting layers {permute_layer}, for {key}----')

        # prep data
        (
            names, labels, cowardin_labels, resource_labels, ajd_labels
         ) = data.make_split_dicts(split_state, verbose=False)

        datasets, dataloaders, data_config = data.get_NAIP_dataloaders(
            names, labels, batch_size, num_cores, 
            normalize=True,
            normalize_path='/home/sgreenhill/repos/WOTUS/new_imagery/4_dl_models/wotus/train/layer_mean_sd',
            n_naip_bands=4, 
            naip=True, nwi=True, 
            nhd=True, nhd_dir='../../../hdruckenmiller/NHD_new/', nhd_prefix='NHD_',
            dem=True, ecoregion=True, nlcd=True, prism=True,
            gnatsgo=True, hq_dist=True,
            district_dummies=True, regrid=False, rule_dummies=True, rule_one_hot=True,
            states=True,
            return_datasets=True, return_config=True, 
            cowardin_labels=cowardin_labels, resource_labels=resource_labels,
            ajd_labels=ajd_labels,
            save=True,
            save_dir=load_path,
            load_saved=True,
            folds=['test'],
            permute_layer=permute_layer
        )

        loss_fn = nn.CrossEntropyLoss()
        metrics = {'accuracy': torch_utils.accuracy}
        save_summary_steps = 1
        targets = ['wotus']

        # evaluate model
        print('Evaluating...')
        optimizer = optim.Adam(model.parameters(), lr=learning_rate, weight_decay=regularization)

        metrics = train.run_model(
            model, optimizer, loss_fn, dataloaders['test'], metrics,
            save_summary_steps, mode='eval', testing=False,
            targets=targets, cuda_device=default_cuda_device
        )
        
        this_acc = (
            np.sum(np.array(metrics['predictions']) == np.array(metrics['labels'])) / \
            len(metrics['predictions'])
        )
        print(f"Accuracy: {this_acc}")
        
        state_permute_accs_blocks[key] = this_acc
    
    permute_accs_by_state[state] = state_permute_accs_blocks

# save results for plotting
savepath = f'{path}/figure_files/figS9_data/permute_accs_blocks_state.p'
with open(savepath, 'wb') as f:
    pickle.dump(permute_accs_by_state, f)

savepath2 = f'{path}/figure_files/figS9_data/test_obs_state.p'
with open(savepath2, 'wb') as f:
    pickle.dump(test_obs_by_state, f)
        
